import os

import ctc.config


def get_command_spec():
    return {
        'f': cd_command,
        'args': [
            {'name': 'dirname'},
        ],
        'special': {'cd': True},
    }


def cd_command(dirname, new_dir_tempfile):

    if new_dir_tempfile is None:
        raise Exception('must specify new_dir_tempfile')

    if dirname == 'code':
        path = ctc.__path__[0]

    elif dirname == 'data':
        path = ctc.config.get_data_dir()

    elif dirname == 'default_data':
        path = ctc.config.get_default_data_dir()

    elif dirname == 'config':
        path = os.path.dirname(ctc.config.get_config_path())

    elif dirname == 'reports':
        path = ctc.config.get_reports_dir()

    else:
        raise Exception('unknown directory: ' + str(dirname))

    with open(new_dir_tempfile, 'w') as f:
        f.write(path)




import tooltime
import toolstr

from ctc import binary
from . import block_crud


async def async_print_block_summary(block, provider=None):

    if not isinstance(block, dict):
        block = await block_crud.async_get_block(block=block, provider=provider)

    full_transactions = len(block['transactions']) > 0 and isinstance(
        block['transactions'][0], dict
    )
    percentiles = [
        0,
        5,
        50,
        95,
        100,
    ]

    if full_transactions:
        gas_prices = [
            transaction['gas_price'] / 1e9
            for transaction in block['transactions']
        ]
        import numpy as np

        gas_percentiles = np.percentile(
            gas_prices,
            percentiles,
        )

    title = 'Block ' + str(block['number'])
    print(title)
    print('─' * len(title))
    print('- timestamp:', block['timestamp'])
    print('- time:', tooltime.timestamp_to_iso(block['timestamp']))
    print('- block_hash:', block['hash'])
    print('- n_transactions:', len(block['transactions']))
    print(
        '- gas used:',
        toolstr.format(block['gas_used']),
        '/',
        toolstr.format(block['gas_limit']),
    )
    if full_transactions:
        percentile_label = (
            '('
            + ', '.join([str(percentile) + '%' for percentile in percentiles])
            + ')'
        )
        gas_percentiles = (
            '('
            + ', '.join(
                [toolstr.format(percentile) for percentile in gas_percentiles]
            )
            + ')'
        )
        print('- gas prices:', percentile_label, '=', gas_percentiles)

    message = block['extra_data']
    try:
        message = binary.convert(message, 'binary').decode()
    except Exception as e:
        if len(message) > 80:
            message = message[:77] + '...'
        else:
            message = message
    print('- message:', message)




from __future__ import annotations

from ctc import evm
from ctc import rpc
from ctc import spec
from . import coracle_spec


async def async_get_tokens_in_pcv(
    block: spec.BlockReference = 'latest',
    wrapper: bool = False,
    provider: spec.ProviderSpec = None,
) -> list[spec.Address]:
    """get list of all tokens in pcv"""

    block = evm.standardize_block_number(block)
    coracle = coracle_spec.get_coracle_address(wrapper=wrapper, block=block)

    return await rpc.async_eth_call(
        to_address=coracle,
        function_name='getTokensInPcv',
        block_number=block,
        provider=provider,
    )


async def async_get_pcv_tokens_symbols(
    tokens: list[spec.ERC20Reference],
    block: spec.BlockNumberReference = 'latest',
    provider: spec.ProviderSpec = None,
):

    non_usd_tokens = list(
        token for token in tokens if token != coracle_spec.usd_token
    )
    symbols = await evm.async_get_erc20s_symbols(
        tokens=non_usd_tokens,
        provider=provider,
        block=block,
    )
    tokens_symbols = dict(zip(non_usd_tokens, symbols))
    if coracle_spec.usd_token in tokens:
        tokens_symbols[coracle_spec.usd_token] = 'USD'

    return tokens_symbols




from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


async def async_eth_get_transaction_count(
    from_address,
    block_number='latest',
    provider=None,
    decode_response=True,
):
    request = rpc_constructors.construct_eth_get_transaction_count(
        from_address=from_address,
        block_number=block_number,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_transaction_count(
        response=response,
        decode_response=decode_response,
    )


async def async_eth_get_transaction_by_hash(
    transaction_hash,
    provider=None,
    decode_response=True,
    snake_case_response=True,
):
    request = rpc_constructors.construct_eth_get_transaction_by_hash(
        transaction_hash=transaction_hash
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_transaction_by_hash(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


async def async_eth_get_transaction_by_block_hash_and_index(
    block_hash,
    transaction_index,
    provider=None,
    decode_response=True,
    snake_case_response=True,
):
    request = (
        rpc_constructors.construct_eth_get_transaction_by_block_hash_and_index(
            block_hash=block_hash,
            transaction_index=transaction_index,
        )
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_transaction_by_block_hash_and_index(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


async def async_eth_get_transaction_by_block_number_and_index(
    block_number,
    transaction_index,
    provider=None,
    decode_response=True,
    snake_case_response=True,
):
    request = rpc_constructors.construct_eth_get_transaction_by_block_number_and_index(
        block_number=block_number,
        transaction_index=transaction_index,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_transaction_by_block_number_and_index(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


async def async_eth_get_transaction_receipt(
    transaction_hash,
    provider=None,
    decode_response=True,
    snake_case_response=True,
):
    request = rpc_constructors.construct_eth_get_transaction_receipt(
        transaction_hash=transaction_hash,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_transaction_receipt(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


async def async_eth_get_block_transaction_count_by_hash(
    block_hash,
    provider=None,
    decode_response=True,
):
    request = rpc_constructors.construct_eth_get_block_transaction_count_by_hash(
        block_hash=block_hash,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_block_transaction_count_by_hash(
        response=response,
        decode_response=decode_response,
    )


async def async_eth_get_block_transaction_count_by_number(
    block_number,
    provider=None,
    decode_response=True,
):
    request = rpc_constructors.construct_eth_get_block_transaction_count_by_number(
        block_number=block_number,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_block_transaction_count_by_number(
        response=response,
        decode_response=decode_response,
    )




import pytest

from ctc import evm


@pytest.mark.asyncio
async def test_fetch_eth_total_supply():
    result = await evm.async_get_eth_total_supply()
    assert isinstance(result, int)


@pytest.mark.asyncio
async def test_feth_eth_balance():
    result = await evm.async_get_eth_balance(
        address='0x00192Fb10dF37c9FB26829eb2CC623cd1BF599E8',
        block=13437523,
        normalize=False,
    )
    assert result == 3463747527330489047936



from __future__ import annotations

import typing
from typing_extensions import TypedDict

from ctc import spec


#
# # types
#


class Timescale(TypedDict):
    window_size: str
    interval_size: str


TimescaleShorthand = str
TimescaleSpec = typing.Union[Timescale, TimescaleShorthand]


Timestamp = int
Timestamps = typing.List[int]

MetricName = str
MetricSeries = typing.List[float]


class TimeData(TypedDict):
    timestamps: list[int]
    block_numbers: list[int]
    n_samples: int
    window_size: str
    interval_size: str
    created_at_timestamp: int


class MetricData(TypedDict, total=False):
    values: typing.Union[list[float], list[int]]
    name: str
    link: str
    units: str


class MetricGroupStrict(TypedDict):
    name: str
    metrics: dict[str, MetricData]


class MetricGroup(MetricGroupStrict, total=False):
    order: list[str]


class AnalyticsPayload(TypedDict):

    # metadata
    version: str
    created_at_timestamp: Timestamp

    # timing data
    n_samples: int
    interval_size: str
    window_size: str
    timestamps: list[Timestamp]
    block_numbers: list[int]

    # metrics
    data: dict[str, MetricGroup]


MetricGroupCreator = typing.Callable[[typing.List[int], bool], MetricGroup]
MetricGroupCreatorCoroutine = typing.Callable[
    [typing.List[int], bool],
    typing.Coroutine[typing.Any, typing.Any, MetricGroup],
]
MultiMetricGroupCreator = typing.Callable[
    [typing.List[int], bool],
    typing.Dict[str, MetricGroup],
]
MultiMetricGroupCreatorCoroutine = typing.Callable[
    [typing.List[int], bool],
    typing.Coroutine[typing.Any, typing.Any, typing.Dict[str, MetricGroup]],
]


#
# # specific values
#

payload_timescales = [
    {'window_size': '1h', 'interval_size': '1 minute'},  # 60 datapoints
    {'window_size': '24h', 'interval_size': '10 minute'},  # 144 datapoints
    {'window_size': '7d', 'interval_size': '1 hour'},  # 168 datapoints
    {'window_size': '30d', 'interval_size': '1 day'},  # 30 datapoints
    {'window_size': '90d', 'interval_size': '1 day'},  # 90 datapoints
]


class DexPoolMetadata(TypedDict, total=False):
    platform: str
    address: spec.ContractAddress
    other_assets: list[str]
    fei_index: int
    event_name: str


dex_pools: dict[str, DexPoolMetadata] = {
    'uniswap_v2__fei_eth': {
        'platform': 'Uniswap V2',
        'address': '0x94b0a3d511b6ecdb17ebf877278ab030acb0a878',
        'other_assets': ['WETH'],
    },
    'uniswap_v2__fei_tribe': {
        'platform': 'Uniswap V2',
        'address': '0x9928e4046d7c6513326ccea028cd3e7a91c7590a',
        'other_assets': ['TRIBE'],
    },
    'uniswap_v3__fei_usdc_1': {
        'platform': 'Uniswap V3',
        'address': '0xdf50fbde8180c8785842c8e316ebe06f542d3443',
        'other_assets': ['USDC'],
    },
    'uniswap_v3__fei_usdc_5': {
        'platform': 'Uniswap V3',
        'address': '0x8c54aa2a32a779e6f6fbea568ad85a19e0109c26',
        'other_assets': ['USDC'],
    },
    'uniswap_v3__fei_usdc_30': {
        'platform': 'Uniswap V3',
        'address': '0x5180545835bd68810fb7e11c7160bb7ea4ae8744',
        'other_assets': ['USDC'],
    },
    'uniswap_v3__fei_dai_5': {
        'platform': 'Uniswap V3',
        'address': '0xbb2e5c2ff298fd96e166f90c8abacaf714df14f8',
        'other_assets': ['DAI'],
    },
    # 'balancer_v2': {
    #     'platform': 'Balancer',
    #     'address': '0xba12222222228d8ba445958a75a0704d566bf2c8',
    #     'other_assets': ['ETH', 'TRIBE'],
    # },
    'curve__fei_3crv': {
        'platform': 'Curve',
        'address': '0x06cb22615BA53E60D67Bf6C341a0fD5E718E1655',
        'other_assets': ['3crv'],
        'fei_index': 0,
        'event_name': 'TokenExchangeUnderlying',
    },
    'curve__d3': {
        'platform': 'Curve',
        'address': '0xBaaa1F5DbA42C3389bDbc2c9D2dE134F5cD0Dc89',
        'other_assets': ['FRAX', 'alUSD'],
        'fei_index': 1,
        'event_name': 'TokenExchange',
    },
    'saddle__d4': {
        'platform': 'Saddle',
        'address': '0xC69DDcd4DFeF25D8a793241834d4cc4b3668EAD6',
        'other_assets': ['alUSD', 'FRAX', 'LUSD'],
        'fei_index': 1,
    },
    'sushiswap__fei_dpi': {
        'platform': 'Sushi',
        'address': '0x8775aE5e83BC5D926b6277579c2B0d40c7D9b528',
        'other_assets': ['DPI'],
    },
}




import typing

from ctc import evm
from ctc import spec
from ctc.toolbox import pd_utils

from .. import chainlink_metadata
from .. import chainlink_spec
from . import feed_datum


async def async_get_full_feed_event_data(
    feed: chainlink_spec._FeedReference,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    normalize: bool = True,
    interpolate: bool = False,
    provider: spec.ProviderSpec = None,
    keep_multiindex: bool = False,
) -> spec.DataFrame:
    """
    TODO: be able to gather data across multiple aggregator changes
    """

    # get feed address
    feed = await chainlink_metadata.async_resolve_feed_address(
        feed, provider=provider
    )

    # get aggregator
    if end_block is None:
        end_block = await evm.async_get_latest_block_number()
    aggregator_address = await chainlink_metadata.async_get_feed_aggregator(
        feed=feed,
        block=end_block,
        provider=provider,
    )

    # get events
    df = await evm.async_get_events(
        contract_address=aggregator_address,
        event_name='AnswerUpdated',
        start_block=start_block,
        end_block=end_block,
        verbose=False,
    )

    # rename columns
    new_columns = {
        'arg__current': 'answer',
        'arg__updatedAt': 'timestamp',
        'arg__roundId': 'round_id',
    }
    df = df.rename(columns=new_columns)
    df = df[['answer', 'timestamp', 'round_id']]

    # normalize
    if normalize:
        decimals = chainlink_metadata.get_feed_decimals(
            feed=feed, provider=provider
        )
        df['answer'] /= 10 ** decimals

    # interpolate
    if interpolate:

        # add initial data
        if start_block < df.index.values[0][0]:
            import pandas as pd

            initial_data = await feed_datum.async_get_feed_datum(
                feed=feed,
                block=start_block,
                provider=provider,
                normalize=normalize,
                fields='full',
            )
            df.index = pd_utils.keep_level(df.index, level='block_number')
            initial_df = pd.DataFrame(initial_data, index=[start_block])
            df = pd.concat([initial_df, df])

        end_block = await evm.async_block_number_to_int(
            end_block, provider=provider
        )
        df = pd_utils.interpolate_dataframe(df, end_index=end_block)

    elif not keep_multiindex:
        df.index = pd_utils.keep_level(df.index, level='block_number')

    return df


async def async_get_answer_feed_event_data(
    feed: chainlink_spec._FeedReference,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    normalize: bool = True,
    interpolate: bool = False,
    provider: spec.ProviderSpec = None,
    keep_multiindex: bool = False,
) -> spec.Series:

    df = await async_get_full_feed_event_data(
        feed=feed,
        start_block=start_block,
        end_block=end_block,
        normalize=normalize,
        interpolate=interpolate,
        provider=provider,
        keep_multiindex=keep_multiindex,
    )

    return df['answer']




import asyncio

from ctc import evm
from ctc import rpc


def get_command_spec():
    return {
        'f': rechunk_command,
        'args': [
            {'name': 'contract_address', 'kwargs': {'nargs': '?'}},
            {'name': 'event', 'kwargs': {'nargs': '?', 'default': None}},
            {'name': '--network'},
            {
                'name': '--all',
                'kwargs': {'action': 'store_true', 'dest': 'all_events'},
            },
            {'name': '--start_block', 'kwargs': {'type': int}},
            {'name': '--end_block', 'kwargs': {'type': int}},
            {'name': '--n_chunks', 'kwargs': {'type': int}},
            {
                'name': '--chunk_target_bytes',
                'kwargs': {'type': float, 'required': True},
            },
            {'name': '--dry', 'kwargs': {'action': 'store_true'}},
            {'name': '--verbose', 'kwargs': {'action': 'store_true'}},
        ],
    }


def rechunk_command(
    contract_address,
    event,
    all_events,
    start_block,
    end_block,
    n_chunks,
    chunk_target_bytes,
    dry,
    verbose,
    network,
):
    coroutine = run(
        contract_address=contract_address,
        event=event,
        all_events=all_events,
        start_block=start_block,
        end_block=end_block,
        n_chunks=n_chunks,
        chunk_target_bytes=chunk_target_bytes,
        dry=dry,
        verbose=verbose,
        network=network,
    )
    asyncio.run(coroutine)


async def run(
    contract_address,
    event,
    all_events,
    start_block,
    end_block,
    n_chunks,
    chunk_target_bytes,
    dry,
    verbose,
    network,
):

    if all_events:

        await evm.async_rechunk_all_events(
            network=network,
            start_block=start_block,
            end_block=end_block,
            chunk_target_bytes=chunk_target_bytes,
            dry=dry,
            verbose=verbose,
        )

    elif contract_address is not None and event is not None:

        if evm.is_event_hash(event):
            event_name = event
            event_hash = None
        else:
            event_name = None
            event_hash = event

        await evm.async_rechunk_events(
            contract_address=contract_address,
            network=network,
            event_name=event_name,
            event_hash=event_hash,
            start_block=start_block,
            end_block=end_block,
            n_chunks=n_chunks,
            chunk_target_bytes=chunk_target_bytes,
            dry=dry,
            verbose=verbose,
        )

    else:
        raise Exception(
            'usage either `ctc rechunk --all` or `ctc rechunk contract_address event_hash`'
        )

    await rpc.async_close_http_session()




"""WIP

Desired Information
- token creation date
- holders
    - top token holders
    - number of holders
- token dex summary on major exchanges
    - volume
    - liquidity and liquidity depth
- token CEX activity and inflow/outflow
- how each of these values change over time
"""




"""
ERC20 lists obtained from https://erc20lists.org/
"""

import typing
from typing_extensions import TypedDict

from ctc import directory
from ctc import spec
from ctc.toolbox import store_utils


class ERC20ListPayload(TypedDict):
    name: str
    timestamp: str
    version: 'ERC20ListVersion'
    keywords: list[str]
    erc20s: list['ERC20ListToken']


class ERC20ListVersion(TypedDict):
    major: int
    minor: int
    patch: int


class ERC20ListToken(TypedDict):
    address: spec.Address
    chainId: int
    name: str
    symbol: str
    decimals: int
    logoURI: str


def import_erc20_list_payload(
    erc20_list_payload: ERC20ListPayload,
    network: spec.NetworkReference,
    label: typing.Optional[str],
    backend: spec.StorageBackend,
    overwrite: bool = False,
) -> None:

    if backend != 'Filesystem':
        raise NotImplementedError('backend ' + str(backend))

    # check that all erc20s are from specified network
    chain_id = directory.get_network_chain_id(network=network)
    for erc20 in erc20_list_payload['erc20s']:
        if erc20['chainId'] != chain_id:
            raise Exception('erc20s from other network in list')

    # get output path
    if label is None:
        label = erc20_list_payload['name']
    output_path = directory.get_erc20_data_path(network=network, label=label)

    # convert data
    data: list[spec.ERC20Metadata] = [
        {
            'name': erc20['name'],
            'address': erc20['address'].lower(),
            'symbol': erc20['symbol'],
            'decimals': erc20['decimals'],
        }
        for erc20 in erc20_list_payload['erc20s']
    ]

    # write data
    store_utils.write_file_data(
        path=output_path, data=data, overwrite=overwrite,
    )

    print('imported data to path', output_path)




import toolstr
import tooltable  # type: ignore

from . import cpmm_trade


def print_pool_summary(
    x_reserves,
    y_reserves,
    lp_total_supply=None,
    x_name=None,
    y_name=None,
    fee_rate=None,
    indent=None,
    depths=None,
):
    # add in +/- 2% depth

    if x_name is None:
        x_name = 'X'
    if y_name is None:
        y_name = 'Y'

    indent = toolstr.indent_to_str(indent)

    print(indent + '- ' + x_name + ' reserves:', toolstr.format(x_reserves, order_of_magnitude=True))
    print(indent + '- ' + y_name + ' reserves:', toolstr.format(y_reserves, order_of_magnitude=True))
    if lp_total_supply is not None:
        print(indent + '- total lp tokens:', toolstr.format(lp_total_supply, order_of_magnitude=True))
    print(
        indent + '-',
        x_name,
        '/',
        y_name + ' price:',
        # '%.6f' % (x_reserves / y_reserves),
        toolstr.format(x_reserves / y_reserves),
    )
    print(
        indent + '-',
        y_name,
        '/',
        x_name + ' price:',
        # '%.6f' % (y_reserves / x_reserves),
        toolstr.format(y_reserves / x_reserves),
    )
    print(indent + '-', x_name + ' / ' + y_name, 'liquidity depth:')
    print()
    print_liquidity_depth(
        x_reserves=x_reserves,
        y_reserves=y_reserves,
        x_name=x_name,
        y_name=y_name,
        fee_rate=fee_rate,
        indent=indent,
        depths=depths,
    )
    print()
    print(indent + '-', y_name + ' / ' + x_name, 'liquidity depth:')
    print()
    print_liquidity_depth(
        x_reserves=y_reserves,
        y_reserves=x_reserves,
        x_name=y_name,
        y_name=x_name,
        fee_rate=fee_rate,
        indent=indent,
        depths=depths,
    )
    print()


def print_liquidity_depth(
    x_reserves,
    y_reserves,
    depths=None,
    x_name=None,
    y_name=None,
    fee_rate=None,
    indent=None,
):

    if x_name is None:
        x_name = 'X'
    if y_name is None:
        y_name = 'Y'

    if depths is None:
        depths = [-0.10, -0.05, -0.02, 0, 0.02, 0.05, 0.10]

    format_str = '{:,.2f}'

    current_x_per_y = x_reserves / y_reserves
    headers = ['depth', 'new price', x_name, y_name]

    trades = []
    for depth in depths:

        trade = []

        # x per y
        if depth == 0:
            trade.append(' 0%')
        else:
            trade.append('%+.0f' % (depth * 100) + '%')

        # new price
        new_x_per_y = type(current_x_per_y)(1 + depth) * current_x_per_y
        trade.append(toolstr.format(new_x_per_y))
        trade[-1] = trade[-1] + ' ' + x_name + ' / ' + y_name

        # buys and sells
        result = cpmm_trade.trade_to_price(
            x_reserves=x_reserves,
            y_reserves=y_reserves,
            new_x_per_y=new_x_per_y,
            fee_rate=fee_rate,
        )
        if depth != 0 and result['x_sold'] > 0:
            trade.append('sell ' + toolstr.format(result['x_sold'], order_of_magnitude=True))
            trade.append(' buy ' + toolstr.format(result['y_bought'], order_of_magnitude=True))
        elif depth != 0 and result['x_sold'] < 0:
            trade.append(' buy ' + toolstr.format(result['x_bought'], order_of_magnitude=True))
            trade.append('sell ' + toolstr.format(result['y_sold'], order_of_magnitude=True))
        else:
            trade.append('     0.00')
            trade.append('     0.00')
        trades.append(trade)

    indent = ' ' * 4 + toolstr.indent_to_str(indent)
    tooltable.print_table(rows=trades, headers=headers, indent=indent)


def print_trade_summary(
    x_name=None,
    y_name=None,
    x_holdings_before=None,
    y_holdings_before=None,
    indent=None,
    **trade_kwargs
):

    if x_name is None:
        x_name = 'X'
    if y_name is None:
        y_name = 'Y'

    trade_summary = summarize_trade(**trade_kwargs)

    x_sold = trade_summary['trade_results']['x_sold']
    y_sold = trade_summary['trade_results']['y_sold']

    indent = toolstr.indent_to_str(indent=indent)

    if x_sold == 0:
        print(indent + 'trade size of 0')
    elif x_sold > 0:
        print(indent + '-', x_name, 'sold:', toolstr.format(x_sold))
        print(indent + '-', y_name, 'bought:', toolstr.format(-y_sold))
        fees = trade_summary['x_fees']
        print(indent + '- fees:', toolstr.format(fees), x_name)
    else:
        print(indent + '-', x_name, 'bought:', toolstr.format(-x_sold))
        print(indent + '-', y_name, 'sold:', toolstr.format(y_sold))
        fees = trade_summary['y_fees']
        print(indent + '- fees:', toolstr.format(fees), y_name)
    print(indent + '- prices:')
    headers = [
        '',
        'P_mean',
        'P_start',
        'P_end',
        'mean slippage',
        'end slippage',
    ]
    rows = []
    row = [
        x_name + ' / ' + y_name,
        toolstr.format(trade_summary['mean_x_per_y']),
        toolstr.format(trade_summary['x_per_y_start']),
        toolstr.format(trade_summary['x_per_y_end']),
        toolstr.format(trade_summary['mean_slippage_x_per_y']),
        toolstr.format(trade_summary['end_slippage_x_per_y']),
    ]
    rows.append(row)
    row = [
        y_name + ' / ' + x_name,
        toolstr.format(trade_summary['mean_y_per_x']),
        toolstr.format(trade_summary['y_per_x_start']),
        toolstr.format(trade_summary['y_per_x_end']),
        toolstr.format(trade_summary['mean_slippage_y_per_x']),
        toolstr.format(trade_summary['end_slippage_y_per_x']),
    ]
    rows.append(row)
    print()
    tooltable.print_table(
        rows=rows, headers=headers, indent='    ' + indent, column_gap_spaces=1
    )

    print()
    print(indent + '- pool reserve sizes:')
    print()
    headers = ['', 'before', 'after', 'change']
    new_x_reserves = trade_summary['trade_results']['new_pool']['x_reserves']
    new_y_reserves = trade_summary['trade_results']['new_pool']['y_reserves']
    x_change = new_x_reserves / trade_kwargs['x_reserves'] - 1
    y_change = new_y_reserves / trade_kwargs['y_reserves'] - 1
    rows = [
        [
            x_name,
            toolstr.format(trade_kwargs['x_reserves']),
            toolstr.format(new_x_reserves),
            toolstr.format(x_change),
        ],
        [
            y_name,
            toolstr.format(trade_kwargs['y_reserves']),
            toolstr.format(new_y_reserves),
            toolstr.format(y_change),
        ],
    ]
    tooltable.print_table(
        rows=rows,
        headers=headers,
        indent='    ' + indent,
        column_gap_spaces=1,
        decimal_places=2,
    )


def summarize_trade(**trade_kwargs):
    """compute y_bought and new pool values when trading x_sold"""

    # compute trade
    results = cpmm_trade.trade(**trade_kwargs)

    # compute mean price
    mean_x_per_y = results['x_sold'] / results['y_bought']
    mean_y_per_x = 1 / mean_x_per_y

    # compute slippage
    x_per_y_start = trade_kwargs['x_reserves'] / trade_kwargs['y_reserves']
    y_per_x_start = 1 / x_per_y_start
    x_per_y_end = (
        results['new_pool']['x_reserves'] / results['new_pool']['y_reserves']
    )
    y_per_x_end = 1 / x_per_y_end
    end_slippage_x_per_y = x_per_y_end / x_per_y_start - 1
    end_slippage_y_per_x = y_per_x_end / y_per_x_start - 1
    mean_slippage_x_per_y = mean_x_per_y / x_per_y_start - 1
    mean_slippage_y_per_x = mean_y_per_x / y_per_x_start - 1

    # compute fees
    if trade_kwargs.get('fee_rate') is None:
        trade_kwargs['fee_rate'] = uniswap_spec.default_trade_fee
    if results['x_sold'] > 0:
        x_fees = results['x_sold'] * trade_kwargs['fee_rate']
        y_fees = 0
    else:
        y_fees = results['y_sold'] * trade_kwargs['fee_rate']
        x_fees = 0

    return {
        'end_slippage_x_per_y': end_slippage_x_per_y,
        'end_slippage_y_per_x': end_slippage_y_per_x,
        'mean_slippage_x_per_y': mean_slippage_x_per_y,
        'mean_slippage_y_per_x': mean_slippage_y_per_x,
        'mean_x_per_y': mean_x_per_y,
        'mean_y_per_x': mean_y_per_x,
        'x_per_y_start': x_per_y_start,
        'y_per_x_start': y_per_x_start,
        'x_per_y_end': x_per_y_end,
        'y_per_x_end': y_per_x_end,
        'x_fees': x_fees,
        'y_fees': y_fees,
        'trade_results': results,
    }




import pytest

from ctc.protocols import balancer_utils


@pytest.mark.asyncio
async def test_get_pool_id():
    pool_id = await balancer_utils.async_get_pool_id(
        '0xede4efcc5492cf41ed3f0109d60bc0543cfad23a'
    )
    assert (
        pool_id
        == '0xede4efcc5492cf41ed3f0109d60bc0543cfad23a0002000000000000000000bb'
    )


@pytest.mark.asyncio
async def test_get_pool_address():
    pool_address = await balancer_utils.async_get_pool_address(
        '0xede4efcc5492cf41ed3f0109d60bc0543cfad23a0002000000000000000000bb'
    )
    assert pool_address == '0xede4efcc5492cf41ed3f0109d60bc0543cfad23a'


@pytest.mark.asyncio
async def test_get_pool_swaps():
    pool_address = '0xede4efcc5492cf41ed3f0109d60bc0543cfad23a'
    swaps = await balancer_utils.async_get_pool_swaps(
        pool_address=pool_address, end_block=13613577
    )
    assert len(swaps) == 5




import typing
from typing_extensions import TypedDict, Literal

from . import address_types
from . import binary_types


BlockHash = binary_types.PrefixHexData
TransactionHash = binary_types.PrefixHexData

block_number_names = ['latest', 'earliest', 'pending']
BlockNumberName = typing.Union[
    Literal['latest'],
    Literal['earliest'],
    Literal['pending'],
]

# anything that can be converted to an int without node querying
RawBlockNumber = typing.Union[typing.SupportsRound, binary_types.HexData]

# an int or block number name
StandardBlockNumber = typing.Union[int, BlockNumberName]

# anything that refers to a block number, raw or standard
BlockNumberReference = typing.Union[RawBlockNumber, StandardBlockNumber]

# any reference to a block
BlockReference = typing.Union[BlockNumberReference, BlockHash]


class BlockRange(TypedDict):
    start_block: StandardBlockNumber
    end_block: StandardBlockNumber
    open_start: bool
    open_end: bool


class BlockSample(TypedDict):
    start_block: StandardBlockNumber
    end_block: StandardBlockNumber
    block_interval: typing.Union[int, None]
    open_start: bool
    open_end: bool


class RawTransaction(TypedDict):
    pass


# use literal definition because 'from' is a python keyword
Transaction = TypedDict(
    'Transaction',
    {
        'hash': TransactionHash,
        'block_hash': BlockHash,
        'block_number': int,
        'chain_id': binary_types.PrefixHexData,
        'from': address_types.Address,
        'gas': int,
        'gas_price': int,
        'input': binary_types.PrefixHexData,
        'nonce': int,
        'r': binary_types.PrefixHexData,
        's': binary_types.PrefixHexData,
        'to': address_types.Address,
        'transaction_index': int,
        'type': binary_types.PrefixHexData,
        'v': int,
        'value': int,
    },
)


class RawBlock(TypedDict):
    pass


class Block(TypedDict):
    number: int
    difficulty: int
    extra_data: binary_types.PrefixHexData
    gas_limit: int
    gas_used: int
    hash: BlockHash
    logs_bloom: binary_types.PrefixHexData
    miner: address_types.Address
    mix_hash: BlockHash
    nonce: binary_types.PrefixHexData
    parent_hash: BlockHash
    receipts_root: binary_types.PrefixHexData
    sha3_uncles: binary_types.PrefixHexData
    size: int
    state_root: binary_types.PrefixHexData
    timestamp: int
    total_difficulty: int
    transactions: typing.Union[
        typing.List[TransactionHash], typing.List[Transaction]
    ]
    transactions_root: binary_types.PrefixHexData
    uncles: typing.List[BlockHash]


class RawLog(TypedDict):
    removed: bool
    logIndex: int
    transactionIndex: int
    transactionHash: TransactionHash
    blockHash: BlockHash
    blockNumber: int
    address: address_types.Address
    data: binary_types.PrefixHexData
    topics: typing.List[binary_types.PrefixHexData]


class PendingRawLog(TypedDict):
    # many log fields are nullable if a log is pending
    removed: bool
    logIndex: typing.Union[None, int]
    transactionIndex: typing.Union[None, int]
    transactionHash: typing.Union[None, TransactionHash]
    blockHash: typing.Union[None, BlockHash]
    blockNumber: typing.Union[None, int]
    address: address_types.Address
    data: binary_types.PrefixHexData
    topics: typing.List[binary_types.PrefixHexData]


NormalizedLog = typing.Dict[str, typing.Any]




from __future__ import annotations

from ctc import spec
from .. import rpc_format
from .. import rpc_request


def construct_eth_new_filter(
    address: spec.BinaryData = None,
    topics: list[spec.BinaryData] = None,
    start_block: spec.BlockSpec = None,
    end_block: spec.BlockSpec = None,
) -> spec.RpcRequest:
    if start_block is not None:
        start_block = rpc_format.encode_block_number(start_block)
    if end_block is not None:
        end_block = rpc_format.encode_block_number(end_block)

    parameters = {
        'address': address,
        'topics': topics,
        'fromBlock': start_block,
        'toBlock': end_block,
    }
    parameters = {k: v for k, v in parameters.items() if v is not None}

    return rpc_request.create('eth_newFilter', [parameters])


def construct_eth_new_block_filter() -> spec.RpcRequest:
    return rpc_request.create('eth_newBlockFilter', [])


def construct_eth_new_pending_transaction_filter() -> spec.RpcRequest:
    return rpc_request.create('eth_newPendingTransactionFilter', [])


def construct_eth_uninstall_filter(
    filter_id: spec.BinaryInteger,
) -> spec.RpcRequest:
    return rpc_request.create('eth_uninstallFilter', [filter_id])


def construct_eth_get_filter_changes(
    filter_id: spec.BinaryInteger,
) -> spec.RpcRequest:
    return rpc_request.create('eth_getFilterChanges', [filter_id])


def construct_eth_get_filter_logs(
    filter_id: spec.BinaryInteger,
) -> spec.RpcRequest:
    return rpc_request.create('eth_getFilterLogs', [filter_id])


def construct_eth_get_logs(
    address: spec.BinaryData = None,
    topics: list[spec.BinaryData] = None,
    start_block: spec.BlockSpec = None,
    end_block: spec.BlockSpec = None,
    block_hash: spec.BinaryData = None,
) -> spec.RpcRequest:
    if start_block is not None:
        start_block = rpc_format.encode_block_number(start_block)
    if end_block is not None:
        end_block = rpc_format.encode_block_number(end_block)

    parameters = {
        'address': address,
        'topics': topics,
        'fromBlock': start_block,
        'toBlock': end_block,
        'blockHash': block_hash,
    }
    parameters = {k: v for k, v in parameters.items() if v is not None}

    return rpc_request.create('eth_getLogs', [parameters])




from ctc import spec
from ctc import binary
from .. import rpc_request
from .. import rpc_format


def construct_eth_get_transaction_count(
    from_address: spec.BinaryData,
    block_number: spec.BlockSpec = 'latest',
) -> spec.RpcRequest:

    block_number = rpc_format.encode_block_number(block_number)
    return rpc_request.create(
        'eth_getTransactionCount',
        [from_address, block_number],
    )


def construct_eth_get_transaction_by_hash(
    transaction_hash: spec.BinaryData,
) -> spec.RpcRequest:
    return rpc_request.create('eth_getTransactionByHash', [transaction_hash])


def construct_eth_get_transaction_by_block_hash_and_index(
    block_hash: spec.BinaryData,
    transaction_index: spec.BinaryData,
) -> spec.RpcRequest:
    transaction_index = binary.convert(transaction_index, 'prefix_hex')

    return rpc_request.create(
        'eth_getTransactionByBlockHashAndIndex',
        [block_hash, transaction_index],
    )


def construct_eth_get_transaction_by_block_number_and_index(
    block_number: spec.BlockSpec,
    transaction_index: spec.BinaryData,
) -> spec.RpcRequest:
    block_number = rpc_format.encode_block_number(block_number)
    transaction_index = binary.convert(transaction_index, 'prefix_hex')

    return rpc_request.create(
        'eth_getTransactionByBlockNumberAndIndex',
        [block_number, transaction_index],
    )


def construct_eth_get_transaction_receipt(
    transaction_hash: spec.BinaryData,
) -> spec.RpcRequest:
    return rpc_request.create(
        'eth_getTransactionReceipt',
        [transaction_hash],
    )


def construct_eth_get_block_transaction_count_by_hash(
    block_hash: spec.BinaryData,
) -> spec.RpcRequest:
    return rpc_request.create(
        'eth_getBlockTransactionCountByHash',
        [block_hash],
    )


def construct_eth_get_block_transaction_count_by_number(
    block_number: spec.BlockSpec,
) -> spec.RpcRequest:

    block_number = rpc_format.encode_block_number(block_number)
    return rpc_request.create(
        'eth_getBlockTransactionCountByNumber',
        [block_number],
    )




from ctc import rpc

from .. import rari_abis


async def async_get_irm_blocks_per_year(interest_rate_model, block='latest'):

    try:
        return await rpc.async_eth_call(
            to_address=interest_rate_model,
            block_number=block,
            function_abi=rari_abis.interest_rate_model_abis['blocksPerYear'],
        )
    except KeyError:
        return await rpc.async_eth_call(
            to_address=rari_abis.contract_examples['InterestRateModel'],
            block_number=block,
            function_abi=rari_abis.interest_rate_model_abis['blocksPerYear'],
        )





import json
import os
import typing

from ctc import config
from ctc import directory
from ctc import spec
from ctc.toolbox import backend_utils


#
# # paths
#


def get_contract_abis_root(
    network: typing.Optional[spec.NetworkReference] = None,
):
    network_name = directory.get_network_name(network)
    return os.path.join(config.get_data_dir(), network_name, 'contract_abis')


def get_contract_abi_dir(
    contract_address: spec.Address,
    network: typing.Optional[spec.NetworkReference] = None,
):
    contract_address = contract_address.lower()
    return os.path.join(
        get_contract_abis_root(network=network), 'contract__' + contract_address
    )


def get_contract_abi_filepath(
    contract_address,
    name,
    network: typing.Optional[spec.NetworkReference] = None,
):
    filename = name + '.json'
    return os.path.join(
        get_contract_abi_dir(contract_address, network=network), filename
    )


#
# # list
#


def list_contract_abi_contracts(network):
    root = get_contract_abis_root(network=network)
    contracts = {}
    for filename in os.listdir(root):
        _, contract_address = filename.split('__')
        contracts[contract_address] = os.path.join(root, filename)
    return contracts


def list_contract_abi_files(
    contract_address,
    network: typing.Optional[spec.NetworkReference] = None,
):
    contract_abi_dir = get_contract_abi_dir(
        contract_address=contract_address, network=network
    )
    paths = []
    if os.path.isdir(contract_abi_dir):
        for filename in sorted(os.listdir(contract_abi_dir)):
            path = os.path.join(contract_abi_dir, filename)
            paths.append(path)
    return paths


def print_contract_abis_summary(
    network: typing.Optional[spec.NetworkReference] = None,
):
    contract_abi_contracts = list_contract_abi_contracts(network=network)
    print('## Contracts with ABI\'s (' + str(len(contract_abi_contracts)) + ')')
    for contract_address, contract_dir in sorted(
        contract_abi_contracts.items()
    ):
        print(
            '-',
            contract_address,
            '(' + str(len(os.listdir(contract_dir))),
            'files)',
        )


#
# # filesystem
#


async def async_save_contract_abi_to_filesystem(
    contract_abi,
    contract_address,
    name=None,
    overwrite=False,
    network: typing.Optional[spec.NetworkReference] = None,
):
    contract_address = contract_address.lower()

    if name is None:
        name = 'contract__' + contract_address
    path = get_contract_abi_filepath(
        contract_address=contract_address,
        name=name,
        network=network,
    )
    if not overwrite and os.path.exists(path):
        raise Exception('path already exists, use overwrite=True')
    os.makedirs(os.path.dirname(path), exist_ok=True)

    print('saving contract_abi to: ' + str(path))
    with open(path, 'w') as f:
        json.dump(contract_abi, f)

    return contract_abi


async def async_save_proxy_contract_abi_to_filesystem(
    contract_address,
    proxy_implementation,
    name='proxy_implementation',
    network: typing.Optional[spec.NetworkReference] = None,
):
    from ctc import evm

    proxy_abi = await evm.async_get_contract_abi(
        contract_address=proxy_implementation,
        network=network,
    )
    await async_save_contract_abi_to_filesystem(
        contract_abi=proxy_abi,
        contract_address=contract_address,
        name=name,
        network=network,
    )


async def async_get_contract_abi_from_filesystem(
    contract_address,
    name=None,
    network: typing.Optional[spec.NetworkReference] = None,
):
    """concatenate all abi files unless a specific name is given"""
    contract_address = contract_address.lower()

    # gather paths
    if name is None:
        paths = list_contract_abi_files(
            contract_address=contract_address,
            network=network,
        )
    else:
        path = get_contract_abi_filepath(
            contract_address=contract_address,
            name=name,
            network=network,
        )
        paths = [path]

    if len(paths) == 0:
        raise backend_utils.DataNotFound(
            'no files found for contract: ' + str(contract_address)
        )
    if len(paths) == 1 and not os.path.isfile(paths[0]):
        raise backend_utils.DataNotFound(
            'file does not exist: ' + str(paths[0])
        )

    # load files
    contract_abi = []
    for path in paths:
        with open(path, 'r') as f:
            contract_abi += json.load(f)

    return contract_abi




import time

import pytest

from ctc import rpc


@pytest.mark.asyncio
async def test_eth_new_filter():
    new_filter = await rpc.async_eth_new_filter()
    assert len(new_filter) > 0 and new_filter.startswith('0x')


@pytest.mark.asyncio
async def test_eth_new_block_filter():
    new_filter = await rpc.async_eth_new_block_filter()
    assert len(new_filter) > 0 and new_filter.startswith('0x')
    # blocks = await rpc.async_eth_get_filter_changes(new_filter)


@pytest.mark.asyncio
async def test_eth_new_pending_transaction_filter():
    new_filter = await rpc.async_eth_new_pending_transaction_filter()
    assert len(new_filter) > 0 and new_filter.startswith('0x')
    # pending_transactions = await rpc.async_eth_get_filter_changes(new_filter)


@pytest.mark.asyncio
async def test_eth_uninstall_filter():
    new_filter = await rpc.async_eth_new_filter()
    await rpc.async_eth_uninstall_filter(new_filter)


@pytest.mark.asyncio
async def test_eth_get_filter_changes():
    new_filter = await rpc.async_eth_new_pending_transaction_filter()
    time.sleep(3)
    pending_transactions = await rpc.async_eth_get_filter_changes(new_filter)
    assert(len(pending_transactions) > 0)


@pytest.mark.asyncio
async def test_eth_get_filter_logs():
    new_filter = await rpc.async_eth_new_filter(
        topics=None,
        start_block=None,
        end_block=None,
    )
    new_filter = await rpc.async_eth_new_filter(
        topics=['0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'],
        start_block=12345677,
        end_block=12345679,
    )
    logs = await rpc.async_eth_get_filter_logs(new_filter)
    assert len(logs) == 375


@pytest.mark.asyncio
async def test_eth_get_logs():
    logs = await rpc.async_eth_get_logs(
        topics=['0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'],
        start_block=12345677,
        end_block=12345679,
    )
    assert len(logs) == 375




"""need to refactor with common backend interface"""

from __future__ import annotations

import ast
import functools
import os
import typing
from typing_extensions import TypedDict

import toolstr

import ctc.config
from ctc import binary
from ctc import directory
from ctc import spec
from ctc import rpc
from ctc.toolbox import backend_utils
from ctc.toolbox import filesystem_utils
from ... import abi_utils
from ... import block_utils
from ... import evm_spec


#
# # paths
#


def get_events_root(
    network: typing.Optional[spec.NetworkReference] = None,
) -> str:
    network_name = directory.get_network_name(network)
    return os.path.join(ctc.config.get_data_dir(), network_name, 'events')


def get_events_contract_dir(
    contract_address: spec.Address,
    network: typing.Optional[spec.NetworkReference] = None,
) -> str:
    contract_address = contract_address.lower()
    return os.path.join(
        get_events_root(network=network), 'contract__' + contract_address
    )


def get_events_event_dir(
    contract_address: spec.Address,
    event_hash: typing.Optional[str] = None,
    event_abi: typing.Optional[spec.EventABI] = None,
    network: typing.Optional[spec.NetworkReference] = None,
) -> str:
    contract_address = contract_address.lower()
    if event_hash is None:
        if event_abi is None:
            raise Exception('must specify more event data')
        event_hash = binary.get_event_hash(event_abi)
    contract_dir = get_events_contract_dir(contract_address, network=network)
    return os.path.join(contract_dir, 'event__' + event_hash)


def get_events_filepath(
    contract_address: spec.Address,
    start_block: int,
    end_block: int,
    event_hash: typing.Optional[str] = None,
    event_abi: typing.Optional[spec.EventABI] = None,
    network: typing.Optional[spec.NetworkReference] = None,
) -> str:

    # create lowercase versions of contract_address and event_hash
    contract_address = contract_address.lower()
    if event_hash is None:
        if event_abi is None:
            raise Exception('must specify more event data')
        event_hash = binary.get_event_hash(event_abi)
    event_hash = event_hash.lower()

    # assemble items into subpath
    subpath = evm_spec.filesystem_layout['evm_events_path'].format(
        contract_address=contract_address,
        event_hash=event_hash,
        start_block=start_block,
        end_block=end_block,
    )

    # add parent directory
    network_name = directory.get_network_name(network)
    return os.path.join(ctc.config.get_data_dir(), network_name, subpath)


#
# # list saved data
#


def list_events_contracts(
    network: typing.Optional[spec.NetworkReference] = None,
) -> list[str]:
    contracts = []
    events_root = get_events_root(network=network)
    if not os.path.isdir(events_root):
        return []
    for contract_dir in os.listdir(events_root):
        contract_address = contract_dir.split('__')[-1]
        contracts.append(contract_address)
    return contracts


_PathEventsResult = typing.Dict[str, typing.Tuple[int, int]]


class _ListEventsResult(TypedDict):
    paths: _PathEventsResult
    block_range: spec.NumpyArray
    block_mask: spec.NumpyArray
    missing_blocks: spec.NumpyArray


def list_contract_events(
    contract_address: spec.Address,
    event_hash: typing.Optional[str] = None,
    event_abi: typing.Optional[spec.EventABI] = None,
    allow_missing_blocks: bool = False,
    network: typing.Optional[spec.NetworkReference] = None,
) -> dict[str, _ListEventsResult]:

    if event_hash is not None:
        query_event_hash = event_hash
    elif event_abi is not None:
        query_event_hash = binary.get_event_hash(event_abi)
    else:
        query_event_hash = None

    # compile path data
    contract_address = contract_address.lower()
    contract_dir = get_events_contract_dir(contract_address, network=network)
    paths: dict[str, _PathEventsResult] = {}
    if not os.path.isdir(contract_dir):
        return {}
    for event_dirname in os.listdir(contract_dir):
        event_dir = os.path.join(contract_dir, event_dirname)
        _, event_hash = event_dirname.split('__')
        if query_event_hash is not None and event_hash != query_event_hash:
            continue
        for filename in os.listdir(event_dir):
            path = os.path.join(event_dir, filename)
            start_block_str, _, end_block_str = os.path.splitext(filename)[
                0
            ].split('__')
            paths.setdefault(event_hash, {})
            paths[event_hash][path] = (int(start_block_str), int(end_block_str))

    import numpy as np

    # create block_range and block_mask
    events: dict[str, _ListEventsResult] = {}
    for event_hash in paths.keys():

        # gather start and end blocks
        start_blocks = []
        end_blocks = []
        for path, (start_block, end_block) in paths[event_hash].items():
            start_blocks.append(start_block)
            end_blocks.append(end_block)

        # create block_range
        min_block = min(start_blocks)
        max_block = max(end_blocks) + 1
        block_range = np.arange(min_block, max_block)

        # create block_mask
        n_blocks = block_range.size
        block_mask = np.zeros(n_blocks)
        for path, (start_block, end_block) in paths[event_hash].items():
            start_index = start_block - min_block
            end_index = n_blocks - (max_block - end_block) + 1
            block_mask[start_index:end_index] += 1
        if (block_mask > 1).sum() > 0:
            raise Exception('overlapping chunks')
        block_mask = block_mask.astype(bool)

        # check if blocks missing
        missing_blocks = block_mask.sum() != n_blocks
        if missing_blocks and not allow_missing_blocks:
            raise Exception('missing blocks')

        events[event_hash] = {
            'paths': paths[event_hash],
            'block_range': block_range,
            'block_mask': block_mask,
            'missing_blocks': missing_blocks,
        }

    return events


def list_events(
    contract_address: str,
    event_hash: typing.Optional[str] = None,
    event_abi: typing.Optional[spec.EventABI] = None,
    allow_missing_blocks: bool = False,
    network: typing.Optional[spec.NetworkReference] = None,
) -> typing.Optional[_ListEventsResult]:

    contract_events = list_contract_events(
        contract_address=contract_address,
        event_hash=event_hash,
        event_abi=event_abi,
        allow_missing_blocks=allow_missing_blocks,
        network=network,
    )

    if len(contract_events) == 1:
        event_hash = list(contract_events.keys())[0]
        return contract_events[event_hash]
    else:
        return None


def list_contracts_events(
    network: typing.Optional[spec.NetworkReference] = None, **kwargs
) -> dict[str, dict[str, _ListEventsResult]]:
    contracts_events = {}
    for contract_address in list_events_contracts(network=network):
        contracts_events[contract_address] = list_contract_events(
            contract_address=contract_address, network=network, **kwargs
        )
    return contracts_events


#
# # disk
#


def print_events_summary() -> None:
    print_events_summary_filesystem()


def print_events_summary_filesystem() -> None:
    contracts_events = list_contracts_events()
    print('## Contracts (' + str(len(contracts_events)) + ')')
    for contract_address in sorted(contracts_events.keys()):
        n_events = len(contracts_events[contract_address])
        print('-', contract_address, '(' + str(n_events) + ' events)')
        contract_events = contracts_events[contract_address]
        for event_hash, event_data in contract_events.items():
            block_range = [
                event_data['block_range'][0],
                event_data['block_range'][-1],
            ]
            n_files = str(len(event_data['paths']))
            dirpath = get_events_event_dir(
                contract_address=contract_address, event_hash=event_hash
            )
            n_bytes = filesystem_utils.get_directory_nbytes_human(dirpath)
            short_hash = event_hash[:6] + '...' + event_hash[-6:]
            print(
                '    -',
                short_hash,
                block_range,
                '(' + n_bytes + 'B in ' + n_files + ' files)',
            )


async def async_save_events_to_filesystem(
    events: spec.DataFrame,
    contract_address: spec.Address,
    start_block: int,
    end_block: int,
    event_abi: typing.Optional[spec.EventABI] = None,
    event_hash: typing.Optional[str] = None,
    event_name: typing.Optional[str] = None,
    overwrite: bool = False,
    verbose: bool = True,
    provider: spec.ProviderSpec = None,
    network: typing.Optional[spec.NetworkReference] = None,
) -> str:

    if network is None:
        provider = rpc.get_provider(provider)
        network = provider['network']
        if network is None:
            raise Exception('could not determine network')
    else:
        network = directory.get_network_name(network)

    contract_address = contract_address.lower()

    if event_abi is None:
        event_abi = await abi_utils.async_get_event_abi(
            contract_address=contract_address,
            event_name=event_name,
            event_hash=event_hash,
            network=network,
        )

    # compute path
    path = get_events_filepath(
        contract_address=contract_address,
        event_hash=event_hash,
        event_abi=event_abi,
        start_block=start_block,
        end_block=end_block,
        network=network,
    )
    if os.path.exists(path) and not overwrite:
        raise Exception('path already exists, use overwrite=True')

    if verbose:
        print('saving events to file:', path)

    # save
    os.makedirs(os.path.dirname(path), exist_ok=True)
    events.to_csv(path)

    return path


async def async_get_events_from_filesystem(
    contract_address: spec.ContractAddress,
    event_hash: typing.Optional[str] = None,
    event_name: typing.Optional[str] = None,
    event_abi: typing.Optional[spec.EventABI] = None,
    verbose: bool = True,
    start_block: typing.Optional[int] = None,
    end_block: typing.Optional[int] = None,
    provider: spec.ProviderSpec = None,
    network: spec.NetworkReference = None,
) -> spec.DataFrame:

    # get network
    if network is None:
        provider = rpc.get_provider(provider)
        network = provider['network']
        if network is None:
            raise Exception('could not determine network')
    else:
        network = directory.get_network_name(network)

    # resolve start_block and end_block
    if start_block is not None:
        start_block = await block_utils.async_block_number_to_int(
            start_block,
            provider=provider,
        )
    if end_block is not None:
        end_block = await block_utils.async_block_number_to_int(
            end_block,
            provider=provider,
        )

    # get event hash
    if event_hash is None:
        if event_abi is None:
            if event_name is None:
                raise Exception('must specify more event information')
            event_abi = await abi_utils.async_get_event_abi(
                contract_address=contract_address,
                event_name=event_name,
                network=network,
            )

        event_hash = binary.get_event_hash(event_abi)

    events = list_contract_events(
        contract_address=contract_address,
        event_abi=event_abi,
        event_hash=event_hash,
        network=network,
    )
    if event_hash not in events or len(events[event_hash]['paths']) == 0:
        raise backend_utils.DataNotFound('no files for event')

    # get paths to load
    paths_to_load = []
    for path, (path_start, path_end) in events[event_hash]['paths'].items():
        if start_block is not None:
            if path_end < start_block:
                continue
        if end_block is not None:
            if end_block < path_start:
                continue
        paths_to_load.append(path)
    if len(paths_to_load) == 0:
        raise backend_utils.DataNotFound('no files for event')

    # print summary
    if verbose:
        if len(paths_to_load) > 0:
            n_files = len(paths_to_load)
            n_bytes_int = sum(os.path.getsize(path) for path in paths_to_load)
            n_bytes = toolstr.format(n_bytes_int / 1024 / 1024) + 'M'
        else:
            n_bytes = '0'
            n_files = 0
        print('loading events (' + n_bytes + 'B', 'across', n_files, 'files)')
        if verbose >= 2:
            for path in paths_to_load:
                print('-', path)

    import pandas as pd

    # load paths
    dfs = []
    for path in paths_to_load:
        df = pd.read_csv(path)
        df = df.set_index(['block_number', 'transaction_index', 'log_index'])
        dfs.append(df)
    df = pd.concat(dfs, axis=0)
    df = df.sort_index()

    # trim unwanted
    if start_block is not None:
        if start_block < events[event_hash]['block_range'][0]:
            raise backend_utils.DataNotFound(
                'start_block outside of filesystem contents'
            )
        mask = df.index.get_level_values(level='block_number') >= start_block
        df = df[mask]
    if end_block is not None:
        if end_block > events[event_hash]['block_range'][-1]:
            raise backend_utils.DataNotFound(
                'end_block outside of filesystem contents'
            )
        mask = df.index.get_level_values(level='block_number') <= end_block
        df = df[mask]

    # convert any bytes
    prefix = 'arg__'
    if event_abi is None:
        event_abi = await abi_utils.async_get_event_abi(
            contract_address=contract_address,
            event_name=event_name,
            event_hash=event_hash,
            network=network,
        )
    for arg in event_abi['inputs']:
        if arg['type'] in ['bytes32']:
            column = prefix + arg['name']
            lam = functools.partial(
                binary.convert,
                output_format='prefix_hex',
            )
            df[column] = df[column].map(ast.literal_eval).map(lam)

    return df




contenttypes = {
    'Call': {
        'jsonrpc': 'Text',
        'method': 'Text',
        'params': 'Map',
        'id': 'Integer',
    },
    'Request': ['Call', ['Call']],
    'Response': {},
    'Provider': {
        'type': ['http', 'websocket'],
        'url': 'Text',
        'session_kwargs': 'Map',
    },
}

#
# # rpc call schema
#

rpc_block_quantities = [
    'baseFeePerGas',
    'number',
    'difficulty',
    'totalDifficulty',
    'size',
    'gasLimit',
    'gasUsed',
    'timestamp',
]
rpc_log_quantities = [
    'logIndex',
    'transactionIndex',
    'blockNumber',
]
rpc_transaction_quantities = [
    'blockNumber',
    'gas',
    'gasPrice',
    'maxFeePerGas',
    'maxPriorityFeePerGas',
    'nonce',
    'transactionIndex',
    'value',
    'v',
]
rpc_transaction_receipt_quantities = [
    'transactionIndex',
    'blockNumber',
    'cumulativeGasUsed',
    'effectiveGasPrice',
    'gasUsed',
    'quantity',
    'status',
]

rpc_result_scalar_quantities = [
    'eth_blockNumber',
    'eth_getUncleCountByBlockHash',
    'eth_getUncleCountByBlockNumber',
    'eth_newFilter',
    'eth_newBlockFilter',
    'eth_newPendingTransactionFilter',
]
rpc_result_map_quantities = {
    'eth_getBlockByHash': rpc_block_quantities,
    'eth_getBlockByNumber': rpc_block_quantities,
    'eth_getUncleByBlockHashAndIndex': rpc_block_quantities,
    'eth_getUncleByBlockNumberAndIndex': rpc_block_quantities,
}
rpc_result_list_map_quantities = {
    'eth_getLogs': rpc_log_quantities,
}


rpc_constructor_batch_inputs = {
    'eth_get_block_by_hash': {'block_hashes': 'block_hash'},
    'eth_get_block_by_number': {'block_numbers': 'block_number'},
    'eth_get_uncle_count_by_block_hash': {'block_hashes': 'block_hash'},
    'eth_get_uncle_count_by_block_number': {'block_numbers': 'block_number'},
    'eth_compile_lll': {'codes': 'code'},
    'eth_compile_solidity': {'codes': 'code'},
    'eth_compile_serpent': {'codes': 'code'},
    'eth_get_logs': {
        'topic_lists': 'topic',
        'addresses': 'addresses',
        'block_hashes': 'block_hash',
    },
    'web3_sha3': {'datas': 'data'},
    'eth_call': {
        'to_addresses': 'to_address',
        'block_numbers': 'block_number',
        'function_parameter_list': 'function_parameters',
    },
    'eth_estimate_gas': {
        'to_addresses': 'to_address',
        'block_numbers': 'block_number',
        'function_parameter_list': 'function_parameters',
    },
    'eth_get_balance': {
        'addresses': 'address',
        'block_numbers': 'block_number',
    },
    'eth_get_storage_at': {
        'addresses': 'address',
        'positions': 'position',
        'block_numbers': 'block_number',
    },
    'eth_get_code': {'addresses': 'address', 'block_numbers': 'block_number'},
    'eth_send_raw_transaction': {'datas': 'data'},
    'eth_get_transaction_count': {
        'addresses': 'address',
        'block_numbers': 'block_number',
    },
    'eth_get_transaction_by_hash': {'transaction_hashes': 'transaction_hash'},
    'eth_get_transaction_by_block_hash_and_index': {
        'block_hashes': 'block_hash',
        'indices': 'index',
    },
    'eth_get_transaction_by_block_number_and_index': {
        'block_numbers': 'block_number',
        'indices': 'index',
    },
    'eth_get_transaction_receipt': {'transaction_hashes': 'transaction_hash'},
    'eth_get_block_transaction_count_by_hash': {'block_hashes': 'block_hash'},
    'eth_get_block_transaction_count_by_number': {
        'block_numbers': 'block_number'
    },
}




from __future__ import annotations

import os

import toolcli

from .. import config_read
from .. import config_spec


def setup_config_path(styles: dict[str, str]) -> tuple[str, bool]:
    # TODO: for each "where" question offer a default path
    # e.g. ~/.config/ctc/config.json

    env_var = config_spec.config_path_env_var
    env_var_value = os.environ.get(env_var)
    default_config_path = config_spec.default_config_path
    old_config_path = config_read.get_config_path(raise_if_dne=False)

    input_kwargs: toolcli.InputFilenameOrDirectoryKwargs = {
        'prompt': 'Where should config be created? Specify a file path or directory path',
        'default_directory': default_config_path,
        'default_filename': 'config.json',
        'style': styles['question'],
    }

    print()
    print()
    toolcli.print('## Config Path', style=styles['header'])

    if os.path.isfile(old_config_path):
        # case: config file already exists

        # print config path source
        if env_var_value not in [None, '']:
            print()
            print(env_var, '=', env_var_value)
        elif old_config_path == default_config_path:
            print()
            toolcli.print(
                'Using default config path',
                toolcli.add_style(default_config_path, style=styles['path']),
            )
        else:
            raise Exception('unknown config path source')
        print()
        print('Config already file exists')

        # ask whether to change config path
        print()
        answer = toolcli.input_yes_or_no(
            'Keep config at this location? ',
            default='yes',
            style=styles['question'],
        )
        if answer:
            new_config_path = old_config_path
            create_new_config = False
        else:
            print()
            new_config_path = toolcli.input_filename_or_directory(
                **input_kwargs
            )
            create_new_config = True

    elif not os.path.isfile(old_config_path):
        # case: config file does not exist

        if env_var_value not in [None, '']:
            # case: set by env var
            print()
            print(env_var, '=', env_var_value)
            print()
            print('Config path is set but config file does not exist')
            print()
            answer = toolcli.input_yes_or_no(
                'Create config at this path? ', style=styles['question']
            )

        else:
            # case: set by default
            print('Environmental variable', env_var, 'not set')
            print()
            answer = toolcli.input_yes_or_no(
                prompt='Use default config location? '
                + default_config_path + ' ',
                default='yes',
                style=styles['question'],
            )

        # ask whether to change config path
        if answer:
            new_config_path = old_config_path
        else:
            print()
            new_config_path = toolcli.input_filename_or_directory(
                **input_kwargs
            )

        create_new_config = True

    # check file extension
    if create_new_config:
        new_config_path = os.path.abspath(new_config_path)
        head, extension = os.path.splitext(new_config_path)
        default_extension = config_spec.allowed_config_filetypes[0]

        # if a directory, add file
        if extension == '':
            print()
            filename = 'config' + default_extension
            new_config_path = os.path.join(new_config_path, filename)
            print('This is a directory, will use', new_config_path)

        elif extension not in config_spec.allowed_config_filetypes:
            print()
            print(
                'File extension',
                extension,
                'not allowed, switching to',
                default_extension,
            )
            print()
            new_config_path = head + default_extension
            print(new_config_path)

    return new_config_path, create_new_config




import asyncio

from ctc import evm
from ctc import rpc


def get_command_spec():
    return {
        'f': transaction_command,
        'args': [
            {'name': 'transaction'},
            {'name': '--sort'},
        ],
    }


def transaction_command(transaction, sort, **kwargs):

    asyncio.run(run(transaction=transaction, sort=sort))


async def run(transaction, sort):

    await evm.async_print_transaction_summary(
        transaction_hash=transaction, sort_logs_by=sort
    )

    await rpc.async_close_http_session()




import asyncio
import json
import time

from ctc import config
from ctc import spec
from .... import address_utils


_last_request = {'time': None}


async def async_get_contract_abi_from_etherscan(contract_address, network=None):
    """fetch contract abi using etherscan"""

    if network is None:
        network = config.get_default_network()
    if network != 'mainnet':
        raise Exception('etherscan is only for mainnnet')

    import aiohttp

    print('fetching abi from etherscan:', contract_address)

    # ratelimit
    cadence = 6
    current_time = time.time()
    if (
        _last_request['time'] is not None
        and current_time < _last_request['time'] + cadence
    ):
        sleep_time = _last_request['time'] + cadence - current_time
        sleep_time = max(0, sleep_time)
        print('sleeping', sleep_time, 'seconds for etherscan ratelimit')
        asyncio.sleep(sleep_time)
    _last_request['time'] = time.time()

    if not address_utils.is_address_str(contract_address):
        raise Exception('not a valid address: ' + str(contract_address))
    url_template = 'http://api.etherscan.io/api?module=contract&action=getabi&address={address}&format=raw'
    session = aiohttp.ClientSession()
    abi_endpoint = url_template.format(address=contract_address)
    async with session.get(abi_endpoint) as response:
        content = await response.text()
    if content == 'Contract source code not verified':
        raise spec.AbiNotFoundException()
    abi = json.loads(content)
    if isinstance(abi, dict) and abi.get('status') == '0':
        raise Exception(
            'could not obtain contract abi from etherscan for '
            + str(contract_address)
        )
    return abi




import pytest

from ctc import directory
from ctc.protocols import fei_utils


@pytest.mark.asyncio
async def test_get_tokens_deposits():

    deposits = await fei_utils.async_get_token_deposits(
        directory.get_erc20_address('DPI'),
        block=13570382,
    )
    assert deposits == (
        '0x60b63ef8f461355207fe1d8102dda938bbd8c3fb',
        '0x9a774a1b1208c323eded05e6daf592e6e59caa55',
        '0x902199755219a9f8209862d09f1891cfb34f59a3',
    )


@pytest.mark.asyncio
async def test_get_deposit_token_balance():

    balance = await fei_utils.async_get_deposit_balance(
        '0x60b63ef8f461355207fe1d8102dda938bbd8c3fb',
        block=13570382,
    )

    assert balance == 154408399224631216120


@pytest.mark.asyncio
async def test_get_tokens_in_pcv():
    tokens = await fei_utils.async_get_tokens_in_pcv(block=13570382, wrapper=False)
    assert tokens == (
        '0x1111111111111111111111111111111111111111',
        '0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2',
        '0x6b175474e89094c44da98b954eedeac495271d0f',
        '0x03ab458634910aad20ef5f1c8ee96f1d6ac54919',
        '0x1494ca1f11d487c2bbe4543e90080aeba4ba3c2b',
        '0x956f47f50a910163d8bf957cf5846d573e7f87ca',
    )


@pytest.mark.asyncio
async def test_token_to_oracle():
    oracle = await fei_utils.async_get_token_oracle(
        '0x6b175474e89094c44da98b954eedeac495271d0f',
        block=13570382,
    )
    assert oracle == '0x231ada12e273edf3fa54cbd90c5c1a73129d5bb9'




import os

from ctc import evm
from ctc import rpc


def get_command_spec():
    return {
        'f': async_address_command,
        'args': [
            {'name': 'address'},
            {'name': ['-v', '--verbose'], 'kwargs': {'action': 'store_true'}},
        ],
    }


async def async_address_command(address, verbose, **kwargs):
    max_width = os.get_terminal_size().columns
    if verbose:
        verbose = 2
    await evm.async_print_address_summary(
        address=address, verbose=verbose, max_width=max_width
    )
    await rpc.async_close_http_session()




import pytest

from ctc import binary


keccak_data_pairs = [
    [
        '0x689268c0ff57a20cd299fa60d3fb374862aff565b20b5f1767906a99e6e09f3ff04ca2b2a5cd22f62941db103c0356df1a8ed20ce322cab2483db67685afd124',
        '0xde77577a693274e6d5e229c326d1ec50b4e62c1d1a40d16e7cacc6a6580757d5'
    ],
]

keccak_text_pairs = [
    [
        'Transfer(address,address,uint256)',
        '0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef',
    ],
]


@pytest.mark.parametrize('input_output', keccak_data_pairs)
def test_keccak(input_output):
    assert binary.keccak(input_output[0], 'prefix_hex') == input_output[1]


@pytest.mark.parametrize('input_output', keccak_text_pairs)
def test_keccak_text(input_output):
    assert binary.keccak_text(input_output[0], 'prefix_hex') == input_output[1]




from .rpc_block_executors_async import *
from .rpc_log_executors_async import *
from .rpc_mining_executors_async import *
from .rpc_node_executors_async import *
from .rpc_state_executors_async import *
from .rpc_submission_executors_async import *
from .rpc_transaction_executors_async import *
from .rpc_whisper_executors_async import *



from __future__ import annotations

import typing


K = typing.TypeVar('K')
V = typing.TypeVar('V')


def list_of_dicts_to_dict_of_lists(
    list_of_dicts: list[dict[K, V]],
) -> dict[K, list[V]]:
    """convert list of dicts into dict of lists

    uses union of all keys across all dicts
    - any dicts that have missing values will use a fill value (e.g. 0 for ints)

    # TODO: would be nice if could work with list of TypedDict's
    """

    # gather all keys
    all_keys = {
        each_key for each_dict in list_of_dicts for each_key in each_dict.keys()
    }

    # get dtype for fill value
    found = False
    for each_dict in list_of_dicts:
        for value in each_dict.values():
            dtype = type(value)
            found = True
            break
        if found:
            break

    # initialize lists
    dict_of_lists: dict[K, list[V]] = {}
    for key in all_keys:
        dict_of_lists[key] = []

    # build lists
    for each_dict in list_of_dicts:
        for key in all_keys:
            dict_of_lists[key].append(each_dict.get(key, dtype()))

    return dict_of_lists


def is_equal(nested_lhs: typing.Mapping, nested_rhs: typing.Mapping) -> bool:
    import json

    return json.dumps(nested_lhs, sort_keys=True) == json.dumps(
        nested_rhs, sort_keys=True
    )




import decimal

import numpy as np
import pandas as pd

from ctc import directory
from ctc import evm
from ctc import spec

from . import pool_metadata
from . import pool_state


async def async_summarize_pool_state(balancer_pool, block='latest'):

    block = await evm.async_block_number_to_int(block)

    pool_tokens_coroutine = await pool_metadata.async_get_pool_tokens(
        pool_address=balancer_pool,
        block=block,
    )
    pool_fees_coroutine = await pool_state.async_get_pool_fees(
        pool_address=balancer_pool,
        block=block,
    )
    pool_weights_coroutine = await pool_state.async_get_pool_weights(
        pool_address=balancer_pool,
        block=block,
    )
    pool_balances_coroutine = await pool_state.async_get_pool_balances(
        pool_address=balancer_pool,
        block=block,
    )

    return {
        'pool_tokens': pool_tokens_coroutine,
        'pool_fees': pool_fees_coroutine,
        'pool_weights': pool_weights_coroutine,
        'pool_balances': pool_balances_coroutine,
        'block': block,
    }


async def async_get_pool_swaps(
    pool_address=None, start_block=None, end_block=None, vault=None
) -> spec.DataFrame:

    if vault is None:
        vault = directory.get_address(name='Vault', label='balancer')

    if start_block is None:
        start_block = directory.get_address_first_block(
            name='Vault', label='balancer'
        )

    swaps = await evm.async_get_events(
        contract_address=vault,
        event_name='Swap',
        start_block=start_block,
        end_block=end_block,
    )

    if pool_address is not None:
        swaps = swaps[swaps['arg__poolId'].str.startswith(pool_address)]

    return swaps


def summarize_pool_swaps(swaps, weights, as_dataframe=True):

    trade_pairs = set()
    for i, row in swaps[['arg__tokenIn', 'arg__tokenOut']].iterrows():
        trade_pairs.add(tuple(row.values))

    pair_data = {}
    for token_in, token_out in trade_pairs:

        mask = (swaps['arg__tokenIn'] == token_in) & (
            swaps['arg__tokenOut'] == token_out
        )
        pair_swaps = swaps[mask]
        pair_weights = weights[mask.values]
        in_amounts = pair_swaps['arg__amountIn'].map(
            decimal.Decimal
        ) / decimal.Decimal('1e18')
        out_amounts = pair_swaps['arg__amountOut'].map(
            decimal.Decimal
        ) / decimal.Decimal('1e18')
        price_in_per_out = in_amounts / out_amounts
        price_out_per_in = out_amounts / in_amounts
        cummulative_in_amount = np.cumsum(in_amounts)
        cummulative_out_amount = np.cumsum(out_amounts)
        cummulative_price_in_per_out = (
            cummulative_in_amount / cummulative_out_amount
        )
        cummulative_price_out_per_in = (
            cummulative_out_amount / cummulative_in_amount
        )

        data = {
            'in_amounts': in_amounts,
            'out_amounts': out_amounts,
            'price_in_per_out': price_in_per_out,
            'price_out_per_in': price_out_per_in,
            'cummulative_in_amount': cummulative_in_amount,
            'cummulative_out_amount': cummulative_out_amount,
            'cummulative_price_in_per_out': cummulative_price_in_per_out,
            'cummulative_price_out_per_in': cummulative_price_out_per_in,
        }

        for c, weight_column in enumerate(pair_weights.columns):
            data['weight_' + str(c)] = pair_weights[weight_column].values

        if as_dataframe:
            data = pd.DataFrame(data)

        pair_data[(token_in, token_out)] = data

    return pair_data




import asyncio

from ctc import rpc
from ctc.protocols import chainlink_utils

from .. import rari_abis
from . import irm_metadata
from . import pool_metadata
from . import token_metadata


async def async_get_ctoken_price(
    ctoken, oracle=None, block='latest', normalize=True
):

    if oracle is None:
        oracle = await _async_get_ctoken_oracle(ctoken=ctoken)

    price = await rpc.async_eth_call(
        to_address=oracle,
        block_number=block,
        function_abi=rari_abis.oracle_abis['getUnderlyingPrice'],
        function_parameters=[ctoken],
    )
    if normalize:
        price /= 1e18
    return price


async def async_get_total_borrowed(ctoken, block='latest'):
    return await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['totalBorrows'],
        empty_token=0,
    )


async def async_get_total_liquidity(ctoken, block='latest'):
    return await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['getCash'],
        empty_token=0,
    )


async def async_get_reserves(ctoken, block='latest'):
    return await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['totalReserves'],
    )


async def async_get_ctoken_utilization(ctoken, block='latest'):
    borrowed_coroutine = async_get_total_borrowed(ctoken, block)
    liquidity_coroutine = async_get_total_liquidity(ctoken, block)

    borrowed = await borrowed_coroutine
    liquidity = await liquidity_coroutine

    total = borrowed + liquidity

    if total == 0:
        return 0
    else:
        return borrowed / total


async def async_get_ctoken_exchange_rate(ctoken, block='latest'):
    return await rpc.async_eth_call(
        to_address=ctoken,
        function_abi=rari_abis.ctoken_function_abis['exchangeRateCurrent'],
        block_number=block,
        empty_token=None,
    )


async def async_get_ctoken_exchange_rate_by_block(ctoken, blocks):
    return await rpc.async_batch_eth_call(
        to_address=ctoken,
        function_abi=rari_abis.ctoken_function_abis['exchangeRateCurrent'],
        block_numbers=blocks,
        empty_token=None,
    )


async def async_get_supply_interest_per_block(
    ctoken, block='latest', normalize=True
):
    result = await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['supplyRatePerBlock'],
    )
    if normalize:
        result /= 1e18
    return result


async def async_get_borrow_interest_per_block(
    ctoken, block='latest', normalize=True
):
    result = await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['borrowRatePerBlock'],
    )
    if normalize:
        result /= 1e18
    return result


async def async_get_supply_apy(ctoken, blocks_per_year=None, block='latest'):
    supply_interest_per_block = await async_get_supply_interest_per_block(
        ctoken=ctoken,
        block=block,
        normalize=True,
    )
    if blocks_per_year is None:
        irm = await token_metadata.async_get_ctoken_irm(ctoken, block=block)
        blocks_per_year = await irm_metadata.async_get_irm_blocks_per_year(
            irm,
            block=block,
        )

    return (1 + supply_interest_per_block) ** blocks_per_year - 1


async def async_get_borrow_apy(ctoken, blocks_per_year=None, block='latest'):

    borrow_interest_per_block = await async_get_borrow_interest_per_block(
        ctoken=ctoken,
        block=block,
        normalize=True,
    )

    if blocks_per_year is None:
        irm = await token_metadata.async_get_ctoken_irm(ctoken, block=block)
        blocks_per_year = await irm_metadata.async_get_irm_blocks_per_year(
            irm,
            block=block,
        )

    return (1 + borrow_interest_per_block) ** blocks_per_year - 1


async def async_get_ctoken_tvl_and_tvb(
    ctoken,
    oracle=None,
    eth_price=None,
    block='latest',
    in_usd=True,
):
    if not in_usd:
        borrowed = asyncio.create_task(
            async_get_total_borrowed(ctoken=ctoken, block=block)
        )
        liquidity = asyncio.create_task(
            async_get_total_liquidity(ctoken=ctoken, block=block)
        )

        borrowed = await borrowed
        borrowed /= 1e18
        liquidity = await liquidity
        liquidity /= 1e18

        return {'tvb': borrowed, 'tvl': borrowed + liquidity}

    if oracle is None:
        oracle = await _async_get_ctoken_oracle(ctoken)

    if eth_price is None:
        eth_price = await chainlink_utils.async_get_eth_price(block=block)

    # send queries
    borrowed = asyncio.create_task(
        async_get_total_borrowed(ctoken=ctoken, block=block)
    )
    liquidity = asyncio.create_task(
        async_get_total_liquidity(ctoken=ctoken, block=block)
    )
    price = asyncio.create_task(
        async_get_ctoken_price(ctoken=ctoken, oracle=oracle, block=block)
    )

    # receive results
    borrowed = await borrowed
    borrowed /= 1e18
    liquidity = await liquidity
    liquidity /= 1e18
    price = await price

    # compute output
    return {
        'tvb': borrowed * price * eth_price,
        'tvl': (borrowed + liquidity) * price * eth_price,
    }


async def _async_get_ctoken_oracle(ctoken):
    comptroller = await token_metadata.async_get_ctoken_comptroller(ctoken)
    oracle = await pool_metadata.async_get_pool_oracle(comptroller)
    return oracle




from __future__ import annotations

import typing

from ctc import spec
from ctc import rpc

from .. import evm_spec
from . import erc20_normalize
from . import erc20_generic


#
# # total supply
#


async def async_get_erc20_total_supply(
    token: spec.ERC20Reference,
    block: spec.BlockNumberReference = 'latest',
    normalize: bool = True,
    provider: spec.ProviderSpec = None,
    **rpc_kwargs,
) -> typing.Union[int, float]:
    """"""

    total_supply = await erc20_generic.async_erc20_eth_call(
        token=token,
        function_name='totalSupply',
        block=block,
        provider=provider,
        **rpc_kwargs,
    )

    if normalize:
        total_supply = await erc20_normalize.async_normalize_erc20_quantity(
            quantity=total_supply, token=token, provider=provider
        )

    return total_supply


async def async_get_erc20s_total_supplies(
    tokens: typing.Sequence[spec.ERC20Reference],
    block: spec.BlockNumberReference = 'latest',
    normalize: bool = True,
    provider: spec.ProviderSpec = None,
    **rpc_kwargs,
) -> typing.Union[list[int], list[float]]:
    """"""

    total_supplies = await erc20_generic.async_erc20s_eth_calls(
        tokens=tokens, function_name='totalSupply', block=block, **rpc_kwargs
    )

    if normalize:
        total_supplies = (
            await erc20_normalize.async_normalize_erc20s_quantities(
                tokens=tokens, quantities=total_supplies, provider=provider
            )
        )

    return total_supplies


async def async_get_erc20_total_supply_by_block(
    token: spec.ERC20Reference,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: bool = True,
    provider: spec.ProviderSpec = None,
    **rpc_kwargs,
) -> typing.Union[list[int], list[float]]:

    total_supplies = await erc20_generic.async_erc20_eth_call_by_block(
        token=token,
        function_name='totalSupply',
        blocks=blocks,
        provider=provider,
        **rpc_kwargs,
    )

    if normalize:
        total_supplies = (
            await erc20_normalize.async_normalize_erc20_quantities_by_block(
                token=token,
                quantities=total_supplies,
                provider=provider,
                blocks=blocks,
            )
        )

    return total_supplies


#
# # balance of
#


async def async_get_erc20_balance_of(
    address: spec.Address,
    token: spec.ERC20Address,
    block: spec.BlockNumberReference = 'latest',
    normalize: bool = True,
    provider: spec.ProviderSpec = None,
    **rpc_kwargs,
) -> typing.Union[int, float]:

    balance = await erc20_generic.async_erc20_eth_call(
        token=token,
        function_name='balanceOf',
        block=block,
        function_parameters=[address],
        provider=provider,
        **rpc_kwargs,
    )

    if normalize:
        balance = await erc20_normalize.async_normalize_erc20_quantity(
            quantity=balance, token=token, provider=provider, block=block
        )

    return balance


async def async_get_erc20_balance_of_addresses(
    addresses: typing.Sequence[spec.Address],
    token: spec.ERC20Address,
    block: spec.BlockNumberReference = 'latest',
    normalize: bool = True,
    provider: spec.ProviderSpec = None,
    **rpc_kwargs,
) -> typing.Union[list[int], list[float]]:

    balances = await rpc.async_batch_eth_call(
        to_address=token,
        block_number=block,
        function_abi=evm_spec.erc20_abis['balanceOf'],
        function_parameter_list=[[address] for address in addresses],
        provider=provider,
        **rpc_kwargs,
    )

    if normalize:
        balances = await erc20_normalize.async_normalize_erc20_quantities(
            quantities=balances, token=token, provider=provider, block=block
        )

    return balances


async def async_get_erc20s_balance_of(
    address: spec.Address,
    tokens: typing.Sequence[spec.ERC20Address],
    block: spec.BlockNumberReference = 'latest',
    normalize: bool = True,
    provider: spec.ProviderSpec = None,
    **rpc_kwargs,
) -> typing.Union[list[int], list[float]]:
    """"""

    balances = await erc20_generic.async_erc20s_eth_calls(
        tokens=tokens,
        function_name='balanceOf',
        block=block,
        function_parameters=[address],
        provider=provider,
        **rpc_kwargs,
    )

    if normalize:
        balances = await erc20_normalize.async_normalize_erc20s_quantities(
            quantities=balances, tokens=tokens, provider=provider, block=block,
        )

    return balances


async def async_get_erc20_balance_of_by_block(
    address: spec.Address,
    token: spec.ERC20Reference,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize=True,
    provider: spec.ProviderSpec = None,
    empty_token=0,
    **rpc_kwargs,
) -> typing.Union[list[int], list[float]]:
    """"""

    balances = await erc20_generic.async_erc20_eth_call_by_block(
        token=token,
        function_name='balanceOf',
        blocks=blocks,
        function_parameters=[address],
        provider=provider,
        empty_token=empty_token,
        **rpc_kwargs,
    )

    if normalize:
        balances = (
            await erc20_normalize.async_normalize_erc20_quantities_by_block(
                quantities=balances,
                token=token,
                provider=provider,
                blocks=blocks,
            )
        )

    return balances


#
# # allowance
#


async def async_get_erc20_allowance(
    token: spec.ERC20Reference,
    address: spec.Address,
    block: spec.BlockNumberReference,
    normalize: bool = True,
    provider: spec.ProviderSpec = None,
) -> typing.Union[int, float]:

    allowance = await erc20_generic.async_erc20_eth_call(
        token=token,
        function_name='allowance',
        block=block,
        function_parameters=[address],
        provider=provider,
    )

    if normalize:
        allowance = await erc20_normalize.async_normalize_erc20_quantity(
            quantity=allowance, token=token, provider=provider, block=block
        )

    return allowance


async def async_get_erc20_allowance_by_block(
    token: spec.ERC20Reference,
    address: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: bool = True,
    provider: spec.ProviderSpec = None,
) -> typing.Union[list[int], list[float]]:

    allowances = await erc20_generic.async_erc20_eth_call_by_block(
        token=token,
        function_name='allowance',
        blocks=blocks,
        function_parameters=[address],
        provider=provider,
    )

    if normalize:
        allowances = (
            await erc20_normalize.async_normalize_erc20_quantities_by_block(
                quantities=allowances,
                token=token,
                provider=provider,
                blocks=blocks,
            )
        )

    return allowances


async def async_get_erc20s_allowances(
    tokens: typing.Sequence[spec.ERC20Reference],
    address: spec.Address,
    block: spec.BlockNumberReference,
    normalize: bool = True,
    provider: spec.ProviderSpec = None,
) -> typing.Union[list[int], list[float]]:

    allowances = await erc20_generic.async_erc20s_eth_calls(
        tokens=tokens,
        function_name='allowance',
        block=block,
        function_parameters=[address],
        provider=provider,
    )

    if normalize:
        allowances = await erc20_normalize.async_normalize_erc20s_quantities(
            quantities=allowances, tokens=tokens, provider=provider, block=block
        )

    return allowances


async def async_get_erc20s_allowances_by_address(
    token: spec.ERC20Reference,
    addresses: typing.Sequence[spec.Address],
    block: spec.BlockNumberReference,
    normalize: bool = True,
    provider: spec.ProviderSpec = None,
):

    allowances = await rpc.async_batch_eth_call(
        to_address=token,
        block_number=block,
        function_abi=evm_spec.erc20_abis['allowance'],
        function_parameter_list=[[address] for address in addresses],
        provider=provider,
    )

    if normalize:
        allowances = await erc20_normalize.async_normalize_erc20_quantities(
            quantities=allowances, token=token, provider=provider, block=block
        )

    return allowances




from ctc import spec
from .. import rpc_request


def construct_web3_client_version() -> spec.RpcRequest:
    return rpc_request.create('web3_clientVersion', [])


def construct_web3_sha3(data: spec.BinaryData) -> spec.RpcRequest:
    return rpc_request.create('web3_sha3', [data])


def construct_net_version() -> spec.RpcRequest:
    return rpc_request.create('net_version', [])


def construct_net_peer_count() -> spec.RpcRequest:
    return rpc_request.create('net_peerCount', [])


def construct_net_listening() -> spec.RpcRequest:
    return rpc_request.create('net_listening', [])


def construct_eth_protocol_version() -> spec.RpcRequest:
    return rpc_request.create('eth_protocolVersion', [])


def construct_eth_syncing() -> spec.RpcRequest:
    return rpc_request.create('eth_syncing', [])




import decimal

from ctc.toolbox import validate_utils


def trade(
    x_reserves,
    y_reserves,
    x_sold=None,
    x_bought=None,
    y_sold=None,
    y_bought=None,
    new_x_reserves=None,
    new_y_reserves=None,
    fee_rate=None,
):
    """perform trade with AMM

    ## Input Requirements
    - all input values must be positive
    - must always specify both x_reserves and y_reserves
    - must specify exactly one of:
        - x_sold
        - x_bought
        - y_sold
        - y_bought
        - new_x_reserves
        - new_y_reserves
    - values in this list can be scalars or numpy arrays

    ## Inputs
    - [see uniswap_spec.py]
    """

    # validate inputs
    if fee_rate is None:
        fee_rate = 0.003
    value = validate_utils._ensure_exactly_one(
        x_sold, x_bought, y_sold, y_bought, new_x_reserves, new_y_reserves
    )
    validate_utils._ensure_non_negative(value)

    kwargs = {
        'x_reserves': x_reserves,
        'y_reserves': y_reserves,
        'fee_rate': fee_rate,
    }
    reverse_kwargs = {
        'y_reserves': x_reserves,
        'x_reserves': y_reserves,
        'fee_rate': fee_rate,
    }

    if x_sold is not None:

        # case: sell x for y, x specified
        x_bought = -x_sold
        y_bought = compute_y_bought_when_x_sold(x_sold=x_sold, **kwargs)
        y_sold = -y_bought

    elif y_sold is not None:

        # case: sell y for x, y specified
        y_bought = -y_sold
        x_bought = compute_y_bought_when_x_sold(x_sold=y_sold, **reverse_kwargs)
        x_sold = -x_bought

    elif x_bought is not None:

        # case: sell y for x, x specified
        x_sold = -x_bought
        y_sold = compute_x_sold_when_y_bought(
            y_bought=x_bought, **reverse_kwargs
        )
        y_bought = -y_sold

    elif y_bought is not None:

        # case: sell y for x, x specified
        y_sold = -y_bought
        x_sold = compute_x_sold_when_y_bought(y_bought=y_bought, **kwargs)
        x_bought = -x_sold

    else:
        raise Exception('could not compute output')

    return {
        'x_bought': x_bought,
        'x_sold': x_sold,
        'y_bought': y_bought,
        'y_sold': y_sold,
        'fee_rate': fee_rate,
        'new_pool': {
            'x_reserves': x_reserves + x_sold,
            'y_reserves': y_reserves + y_sold,
        },
    }


def trade_to_target_reserves(
    x_reserves,
    y_reserves,
    new_x_reserves=None,
    new_y_reserves=None,
    fee_rate=None,
):
    """compute trade required to reach specific target token reserve amounts"""

    kwargs = {
        'x_reserves': x_reserves,
        'y_reserves': y_reserves,
        'fee_rate': fee_rate,
    }

    # convert reserve targets to bought or sold amounts
    if new_x_reserves is not None:
        if validate_utils._ensure_positive(
            x_reserves - new_x_reserves, error=False
        ):
            x_bought = x_reserves - new_x_reserves
            return trade(x_bought=x_bought, **kwargs)
        else:
            x_sold = new_x_reserves - x_reserves
            return trade(x_sold=x_sold, **kwargs)
    elif new_y_reserves is not None:
        if validate_utils._ensure_positive(
            y_reserves - new_y_reserves, error=False
        ):
            y_bought = y_reserves - new_y_reserves
            return trade(y_bought=y_bought, **kwargs)
        else:
            y_sold = new_y_reserves - y_reserves
            return trade(y_sold=y_sold, **kwargs)
    else:
        raise Exception('specify either new_x_reserves or new_y_reserves')


def trade_to_price(
    x_reserves,
    y_reserves,
    new_x_per_y=None,
    new_y_per_x=None,
    fee_rate=None,
):
    """compute trade required to reach specific price"""

    kwargs = {
        'x_reserves': x_reserves,
        'y_reserves': y_reserves,
        'fee_rate': fee_rate,
    }
    reverse_kwargs = {
        'x_reserves': y_reserves,
        'y_reserves': x_reserves,
        'fee_rate': fee_rate,
    }

    validate_utils._ensure_exactly_one(new_x_per_y, new_y_per_x)

    # convert prices to x per y
    if new_y_per_x is not None:
        new_x_per_y = new_y_per_x ** -1

    # compute trades
    if new_x_per_y >= x_reserves / y_reserves:
        # case: sell x to increase x per y
        x_sold = compute_x_sold_to_reach_price(
            new_x_per_y=new_x_per_y, **kwargs
        )
        return trade(x_sold=x_sold, **kwargs)
    else:
        # case: sell y to decrease x per y
        y_sold = compute_x_sold_to_reach_price(
            new_x_per_y=(new_x_per_y ** -1), **reverse_kwargs
        )
        return trade(y_sold=y_sold, **kwargs)


def compute_x_sold_to_reach_price(
    x_reserves,
    y_reserves,
    new_x_per_y,
    fee_rate=None,
):
    """use quadratic formula to find trade size needed to reach new price

    - see wolframalpha.com/input/?i=g+x%5E2+%2B+%281+%2B+g%29+x+%2B+C+%3D+0
    """
    if fee_rate is None:
        fee_rate = 0.003
    gamma = 1 - fee_rate
    C = 1 - new_x_per_y * y_reserves / x_reserves
    alpha = (gamma + 1) ** 2 - 4 * C * gamma
    if isinstance(gamma, decimal.Decimal):
        alpha = alpha.sqrt()
    else:
        alpha = alpha ** 0.5
    alpha = alpha - gamma - 1
    alpha = alpha / 2 / gamma
    x_sold = alpha * x_reserves
    return x_sold


def compute_y_bought_when_x_sold(x_sold, x_reserves, y_reserves, fee_rate=None):
    """compute amount of y bought when selling x_sold amount of x"""
    if fee_rate is None:
        fee_rate = 0.003
    validate_utils._ensure_non_negative(x_sold)
    alpha = x_sold / x_reserves
    gamma = 1 - fee_rate
    y_bought = alpha * gamma / (1 + alpha * gamma) * y_reserves
    return y_bought


def compute_x_sold_when_y_bought(
    y_bought,
    x_reserves,
    y_reserves,
    fee_rate=None,
):
    """compute amount of x that must be sold to buy y_bought amount of y"""
    if fee_rate is None:
        fee_rate = 0.003
    validate_utils._ensure_non_negative(y_bought)
    beta = y_bought / y_reserves
    gamma = 1 - fee_rate
    x_sold = beta / (1 - beta) / gamma * x_reserves
    return x_sold





from .. import rpc_request


def construct_trace_transaction(transaction_hash):
    return rpc_request.create('trace_transaction', [transaction_hash])


def digest_trace_transaction(response):
    return response


async def async_trace_transaction(transaction_hash, provider=None):
    request = construct_trace_transaction(transaction_hash=transaction_hash)
    response = await rpc_request.async_send(request, provider=provider)
    return digest_trace_transaction(response=response)




from __future__ import annotations

import re
import typing

from ctc import binary
from ctc import spec


T = typing.TypeVar('T')


def encode_block_number(block: spec.BlockSpec) -> str:

    if isinstance(block, str) and block in ['latest', 'earliest', 'pending']:
        return block
    else:
        # python3.7 compatibliity
        supports_int = hasattr(block, '__int__')
        # supports_int = isinstance(block, typing.SupportsInt)

        if supports_int:
            block = int(block)
        return binary.convert(block, 'prefix_hex')


def decode_response(
    response: dict[str, typing.Union[int, str]],
    quantities: typing.Optional[list[str]] = None,
) -> dict:
    if quantities is None:
        quantities = []
    decoded = {}
    for key, value in response.items():
        if key in quantities:
            value = binary.convert(value, 'integer')
        decoded[key] = value
    return decoded


def keys_to_snake_case(map: dict[str, T]) -> dict[str, T]:
    """

    not for general usage beyond the keys that appear in the EVM rpc spec

    adapted from https://stackoverflow.com/a/1176023
    """
    return {camel_case_to_snake_case(key): value for key, value in map.items()}


def camel_case_to_snake_case(text: str) -> str:
    """

    not for general usage beyond the keys that appear in the EVM rpc spec

    adapted from https://stackoverflow.com/a/1176023
    """
    return re.sub(r'(?<!^)(?=[A-Z])', '_', text).lower()




from __future__ import annotations

import sys

import toolcli

from ctc import spec
from .. import config_read


def ensure_valid(styles: dict[str, str]) -> None:
    if config_read.config_path_exists() and not config_read.config_is_valid():
        print()
        print()
        toolcli.print('## Current Config Validation', style=styles['header'])
        print()
        print('Config file already exists but is invalid')
        print()
        print('Config should conform to the following spec:')
        print()
        toolcli.print('{', style=styles['quote'])
        for key, value in spec.ConfigSpec.__annotations__.items():
            toolcli.print(
                '    ' + key + ':',
                value,
                highlight=False,
                style=styles['quote'],
            )
        toolcli.print('}', style=styles['quote'])
        print()
        config_path = config_read.get_config_path(raise_if_dne=False)
        toolcli.print(
            'Spec is violated in current config file:',
            toolcli.add_style(config_path, styles['path']),
        )
        print()
        choices = ['Create new config from scratch', 'Manually repair config']
        answer = toolcli.input_number_choice(
            prompt='How do you want to proceed?',
            choices=choices,
            style=styles['question'],
        )
        if answer == 1:
            print(
                'You should manually edit the file and then restart ctc setup'
            )
            sys.exit()




from .directory_metadata import *
from .irm_metadata import *
from .pool_metadata import *
from .pool_state import *
from .pool_summary import *
from .token_metadata import *
from .token_state import *




from ctc import spec
from . import rpc_batch_utils


def batch_construct_eth_accounts(**constructor_kwargs) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_accounts', **constructor_kwargs
    )


def batch_construct_eth_block_number(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_block_number', **constructor_kwargs
    )


def batch_construct_eth_call(**constructor_kwargs) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_call', **constructor_kwargs
    )


def batch_construct_eth_coinbase(**constructor_kwargs) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_coinbase', **constructor_kwargs
    )


def batch_construct_eth_compile_lll(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_compile_lll', **constructor_kwargs
    )


def batch_construct_eth_compile_serpent(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_compile_serpent', **constructor_kwargs
    )


def batch_construct_eth_compile_solidity(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_compile_solidity', **constructor_kwargs
    )


def batch_construct_eth_estimate_gas(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_estimate_gas', **constructor_kwargs
    )


def batch_construct_eth_gas_price(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_gas_price', **constructor_kwargs
    )


def batch_construct_eth_get_balance(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_balance', **constructor_kwargs
    )


def batch_construct_eth_get_block_by_hash(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_block_by_hash', **constructor_kwargs
    )


def batch_construct_eth_get_block_by_number(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_block_by_number', **constructor_kwargs
    )


def batch_construct_eth_get_block_transaction_count_by_hash(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_block_transaction_count_by_hash', **constructor_kwargs
    )


def batch_construct_eth_get_block_transaction_count_by_number(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_block_transaction_count_by_number', **constructor_kwargs
    )


def batch_construct_eth_get_code(**constructor_kwargs) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_code', **constructor_kwargs
    )


def batch_construct_eth_get_compilers(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_compilers', **constructor_kwargs
    )


def batch_construct_eth_get_filter_changes(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_filter_changes', **constructor_kwargs
    )


def batch_construct_eth_get_filter_logs(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_filter_logs', **constructor_kwargs
    )


def batch_construct_eth_get_logs(**constructor_kwargs) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_logs', **constructor_kwargs
    )


def batch_construct_eth_get_storage_at(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_storage_at', **constructor_kwargs
    )


def batch_construct_eth_get_transaction_by_block_hash_and_index(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_transaction_by_block_hash_and_index',
        **constructor_kwargs,
    )


def batch_construct_eth_get_transaction_by_block_number_and_index(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_transaction_by_block_number_and_index',
        **constructor_kwargs,
    )


def batch_construct_eth_get_transaction_by_hash(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_transaction_by_hash', **constructor_kwargs
    )


def batch_construct_eth_get_transaction_count(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_transaction_count', **constructor_kwargs
    )


def batch_construct_eth_get_transaction_receipt(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_transaction_receipt', **constructor_kwargs
    )


def batch_construct_eth_get_uncle_by_block_hash_and_index(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_uncle_by_block_hash_and_index', **constructor_kwargs
    )


def batch_construct_eth_get_uncle_by_block_number_and_index(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_uncle_by_block_number_and_index', **constructor_kwargs
    )


def batch_construct_eth_get_uncle_count_by_block_hash(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_uncle_count_by_block_hash', **constructor_kwargs
    )


def batch_construct_eth_get_uncle_count_by_block_number(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_uncle_count_by_block_number', **constructor_kwargs
    )


def batch_construct_eth_get_work(**constructor_kwargs) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_get_work', **constructor_kwargs
    )


def batch_construct_eth_hashrate(**constructor_kwargs) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_hashrate', **constructor_kwargs
    )


def batch_construct_eth_mining(**constructor_kwargs) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_mining', **constructor_kwargs
    )


def batch_construct_eth_new_block_filter(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_new_block_filter', **constructor_kwargs
    )


def batch_construct_eth_new_filter(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_new_filter', **constructor_kwargs
    )


def batch_construct_eth_new_pending_transaction_filter(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_new_pending_transaction_filter', **constructor_kwargs
    )


def batch_construct_eth_protocol_version(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_protocol_version', **constructor_kwargs
    )


def batch_construct_eth_send_raw_transaction(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_send_raw_transaction', **constructor_kwargs
    )


def batch_construct_eth_send_transaction(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_send_transaction', **constructor_kwargs
    )


def batch_construct_eth_sign(**constructor_kwargs) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_sign', **constructor_kwargs
    )


def batch_construct_eth_sign_transaction(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_sign_transaction', **constructor_kwargs
    )


def batch_construct_eth_submit_hashrate(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_submit_hashrate', **constructor_kwargs
    )


def batch_construct_eth_submit_work(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_submit_work', **constructor_kwargs
    )


def batch_construct_eth_syncing(**constructor_kwargs) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_syncing', **constructor_kwargs
    )


def batch_construct_eth_uninstall_filter(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='eth_uninstall_filter', **constructor_kwargs
    )


def batch_construct_net_listening(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='net_listening', **constructor_kwargs
    )


def batch_construct_net_peer_count(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='net_peer_count', **constructor_kwargs
    )


def batch_construct_net_version(**constructor_kwargs) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='net_version', **constructor_kwargs
    )


def batch_construct_shh_add_to_group(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='shh_add_to_group', **constructor_kwargs
    )


def batch_construct_shh_get_filter_changes(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='shh_get_filter_changes', **constructor_kwargs
    )


def batch_construct_shh_get_messages(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='shh_get_messages', **constructor_kwargs
    )


def batch_construct_shh_has_identity(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='shh_has_identity', **constructor_kwargs
    )


def batch_construct_shh_new_filter(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='shh_new_filter', **constructor_kwargs
    )


def batch_construct_shh_new_group(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='shh_new_group', **constructor_kwargs
    )


def batch_construct_shh_new_identity(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='shh_new_identity', **constructor_kwargs
    )


def batch_construct_shh_post(**constructor_kwargs) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='shh_post', **constructor_kwargs
    )


def batch_construct_shh_uninstall_filter(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='shh_uninstall_filter', **constructor_kwargs
    )


def batch_construct_shh_version(**constructor_kwargs) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='shh_version', **constructor_kwargs
    )


def batch_construct_web3_client_version(
    **constructor_kwargs,
) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='web3_client_version', **constructor_kwargs
    )


def batch_construct_web3_sha3(**constructor_kwargs) -> spec.RpcPluralRequest:
    return rpc_batch_utils.batch_construct(
        method='web3_sha3', **constructor_kwargs
    )




from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import pandas

    from typing_extensions import TypeGuard


def is_dataframe(candidate) -> TypeGuard['pandas.core.frame.DataFrame']:
    import pandas as pd
    return isinstance(candidate, pd.DataFrame)




from ctc import spec

from . import config_path_setup
from . import data_root_setup
from . import final_setup
from . import network_setup
from . import validation_setup


styles = {
    'header': '#ce93f9 bold',
    'path': '#b9f29f bold',
    'question': '#8be9fd',
    'quote': '#f1fa8c',
}


def setup_ctc() -> None:

    # print intro
    print('ctc initializing...')
    print()
    print('This process will make sure each of the following is completed:')
    print('- setup config path')
    print('- setup data directory')
    print('- setup networks and providers')
    print()
    print('Each step can be skipped depending on what you need')

    # ensure file is valid
    validation_setup.ensure_valid(styles=styles)

    # collect new config file data
    (
        config_path,
        create_because_config_path,
    ) = config_path_setup.setup_config_path(styles=styles)
    data_root, create_because_data_root = data_root_setup.setup_data_root(styles=styles)
    network_data, create_because_networks = network_setup.setup_networks(
        styles=styles
    )

    # TODO: prompt whether to keep any other old config settings
    pass

    # create new config file if need be
    create_new_config = any(
        [
            create_because_config_path,
            create_because_data_root,
            create_because_networks,
        ]
    )

    config: spec.ConfigSpec = {
        'version': '0.1.0',
        'data_dir': data_root,
        'networks': network_data['networks'],
        'providers': network_data['providers'],
        'network_defaults': network_data['network_defaults'],
    }

    final_setup.finalize_setup(
        create_new_config=create_new_config,
        config_path=config_path,
        config=config,
        styles=styles,
    )




from ctc import directory
from ctc import evm

from . import pool_summary


async def plot_lbp_summary(
    swaps,
    weights,
    pool_name,
    pool_tokens,
    pool_address,
    price_range=None,
    premium_range=None,
    oracle_data=None,
):

    summary = pool_summary.summarize_pool_swaps(
        swaps=swaps, weights=weights
    )

    for pair in summary.keys():

        blocks = summary[pair].index.get_level_values('block_number')

        token_0, token_1 = await evm.async_get_erc20s_symbols(pool_tokens)
        in_address, out_address = pair
        if in_address == pool_tokens[0]:
            in_token = token_0
            out_token = token_1
            in_weights = summary[pair]['weight_0']
            out_weights = summary[pair]['weight_1']
        else:
            out_token = token_0
            in_token = token_1
            out_weights = summary[pair]['weight_0']
            in_weights = summary[pair]['weight_1']

        title = pool_name + ', in=' + in_token + ', out=' + out_token

        if oracle_data is not None:
            min_block = swaps.index[0][0]
            max_block = swaps.index[-1][0]
            oracle_mask = (oracle_data.index >= min_block) & (
                oracle_data.index <= max_block
            )
            oracle_data = oracle_data[oracle_mask]
            ys = [
                {
                    'y': oracle_data.values,
                    'x': oracle_data.index.values,
                    'y_kwargs': {'label': 'oracle'},
                },
            ]

            premium = []
            for index, row in summary[pair].iterrows():
                block_number = index[0]
                oracle_price = oracle_data[block_number]
                actual_price = row['price_out_per_in']
                premium.append(float(actual_price) / oracle_price - 1)
        else:
            ys = None
            premium = []

        y_kwargs = {'marker': '.', 'markersize': 10, 'linewidth': 1}
        plot_data = {
            'subplot_height': 5,
            'common': {
                'x': blocks,
                'name_position': 'ylabel',
                'tickgrid': True,
                'y_kwargs': y_kwargs,
            },
            'plots': {
                'price': {
                    'title': title,
                    'name': 'price\n(' + out_token + ' / ' + in_token + ')',
                    'y': summary[pair]['price_out_per_in'],
                    'y_kwargs': dict(label='actual', **y_kwargs),
                    'ys': ys,
                    'ylim': price_range,
                    # 'legend_kwargs': {'loc': 'lower left'},
                },
                'premium': {
                    'name': 'price premium',
                    'y': premium,
                    'ylim': premium_range,
                },
                'weight': {
                    'name': out_token + ' Weights',
                    'y': out_weights,
                },
                'token0 Sold': {
                    'name': in_token + ' Amount',
                    'y': summary[pair]['in_amounts'],
                },
                'Cummulative token0 Bought': {
                    'name': 'Cummulative\n' + in_token + ' Bought',
                    'y': summary[pair]['cummulative_in_amount'],
                },
                'Cummulative token1 Sold': {
                    'name': 'Cummulative\n' + out_token + 'Sold',
                    'y': summary[pair]['cummulative_out_amount'],
                },
                'Volume Weighted Price': {
                    'name': 'Cummulative\nVolume Weighted\nPrice',
                    'y': summary[pair]['cummulative_price_out_per_in'],
                    'xlabel': 'block',
                    'ylim': price_range,
                },
            },
        }

        import toolplot  # type: ignore
        toolplot.plot_subplots(plot_data)




from __future__ import annotations

import typing

from ctc import spec

from . import erc20_metadata


async def async_normalize_erc20_quantity(
    quantity: typing.SupportsFloat,
    token: typing.Optional[spec.ERC20Address],
    provider: spec.ProviderSpec = None,
    decimals: typing.Optional[typing.SupportsInt] = None,
    block: spec.BlockNumberReference = 'latest',
) -> float:
    """convert raw erc20 quantity by adjusting radix by (10 ** decimals)"""

    if quantity == 0:
        return 0

    # get decimals
    if decimals is None:
        if token is None:
            raise Exception('must specify token or decimals')
        decimals = await erc20_metadata.async_get_erc20_decimals(
            token,
            provider=provider,
            block=block,
        )
    else:
        decimals = int(decimals)

    # normalize
    return float(quantity) / (10 ** decimals)


async def async_normalize_erc20_quantities(
    quantities: typing.Sequence[typing.SupportsInt],
    token: spec.ERC20Address,
    provider: spec.ProviderSpec = None,
    decimals: typing.Optional[typing.SupportsInt] = None,
    block: spec.BlockNumberReference = 'latest',
) -> list[float]:

    if all(quantity == 0 for quantity in quantities):
        return quantities

    if decimals is None:
        if token is None:
            raise Exception('must specify token or decimals')
        decimals = await erc20_metadata.async_get_erc20_decimals(
            token=token,
            block=block,
            provider=provider,
        )
    else:
        decimals = int(decimals)

    return [quantity / (10 ** decimals) for quantity in quantities]


async def async_normalize_erc20s_quantities(
    quantities: typing.Sequence[typing.SupportsInt],
    tokens: typing.Optional[typing.Sequence[spec.ERC20Address]] = None,
    decimals: typing.Optional[typing.Sequence[typing.SupportsInt]] = None,
    block: spec.BlockNumberReference = 'latest',
    provider: spec.ProviderSpec = None,
) -> list[float]:

    # take subset of non zero values
    mask = [quantity != 0 for quantity in quantities]
    any_zero = not all(mask)
    if any_zero:
        old_quantities = quantities
        quantities = [
            quantity for quantity, nonzero in zip(quantities, mask) if nonzero
        ]
        if tokens is not None:
            tokens = [token for token, nonzero in zip(tokens, mask) if nonzero]
        if decimals is not None:
            decimals = [
                decimal for decimal, nonzero in zip(decimals, mask) if nonzero
            ]

    if decimals is None:
        if tokens is None:
            raise Exception('must specify tokens or decimals')
        use_decimals = await erc20_metadata.async_get_erc20s_decimals(
            tokens=tokens,
            block=block,
            provider=provider,
        )
    else:
        use_decimals = [int(decimal) for decimal in decimals]

    if len(use_decimals) != len(quantities):
        raise Exception('number of quantities must match number of decimals')

    # put back in zero values
    if any_zero:
        quantities = old_quantities
        new_use_decimals = []
        use_decimals_iterator = iter(use_decimals)
        for nonzero in mask:
            if nonzero:
                new_use_decimals.append(next(use_decimals_iterator))
            else:
                new_use_decimals.append(1)
        use_decimals = new_use_decimals

    return [
        quantity / (10 ** decimal)
        for quantity, decimal in zip(quantities, use_decimals)
    ]


async def async_normalize_erc20_quantities_by_block(
    quantities: typing.Sequence[typing.SupportsInt],
    blocks: typing.Iterable[spec.BlockNumberReference],
    token: typing.Optional[spec.ERC20Address] = None,
    decimals: typing.Optional[list[typing.SupportsInt]] = None,
    provider: spec.ProviderSpec = None,
) -> list[float]:

    # take subset of non zero values
    mask = [quantity != 0 for quantity in quantities]
    any_zero = not all(mask)
    if any_zero:
        old_quantities = quantities
        quantities = [
            quantity for quantity, nonzero in zip(quantities, mask) if nonzero
        ]
        blocks = [
            block for block, nonzero in zip(blocks, mask) if nonzero
        ]
        if decimals is not None:
            decimals = [
                decimal for decimal, nonzero in zip(decimals, mask) if nonzero
            ]

    if decimals is None:
        if token is None:
            raise Exception('must specify token or decimals')
        use_decimals = await erc20_metadata.async_get_erc20_decimals_by_block(
            token=token,
            blocks=blocks,
            provider=provider,
        )
    else:
        use_decimals = [int(decimal) for decimal in decimals]

    if len(use_decimals) != len(quantities):
        raise Exception('number of quantities must match number of decimals')

    if any_zero:
        quantities = old_quantities
        new_use_decimals = []
        use_decimals_iterator = iter(use_decimals)
        for nonzero in mask:
            if nonzero:
                new_use_decimals.append(next(use_decimals_iterator))
            else:
                new_use_decimals.append(1)
        use_decimals = new_use_decimals

    return [
        quantity / (10 ** decimal)
        for quantity, decimal in zip(quantities, use_decimals)
    ]




import copy
import json
import os

import numpy as np
import pandas
import tooltime


vote_data_file_template = '{name}__{datatype}__{timestamp}.json'

datatypes = [
    'choices',
    'scores',
]

proposal_hashes = {
    'FIP_1': 'QmWYNqr9Rcn8QFiZYyRqjQno3jXTToy3vM4bVKJkuntvoz',
}

url_templates = {
    'choices': 'https://hub.snapshot.org/api/fei.eth/proposal/{proposal_hash}',
}

choice_names = {
    'FIP_1': {1: 'no change', 2: '$1.00 / FEI', 3: '$0.90 / FEI'},
}


#
# # internet requests
#


def request_raw_vote_data(name, save=True, overwrite=False, verbose=True):

    # create url
    url = url_templates['choices'].format(proposal_hash=proposal_hashes[name])

    if verbose:
        print('requesting proposal', name)

    # request data
    import requests
    response = requests.get(url=url)
    data = response.json()

    # save data
    if save:
        timestamp = tooltime.create_timestamp_label()
        path = get_vote_data_path(
            name=name, datatype='choices', timestamp=timestamp
        )
        if os.path.isfile(path) and not overwrite:
            raise Exception('path already exists: ' + str(path))
        with open(path, 'w') as f:
            json.dump(data, f)

        if verbose:
            print('saving to:', path)

    return data


#
# # data paths
#


# def get_vote_data_dir():
#     data_dir = config_utils.get_config()['data_root']
#     return os.path.join(data_dir, 'market/votes')


def get_vote_data_path(name, datatype, timestamp):
    filename = vote_data_file_template.format(
        name=name, datatype=datatype, timestamp=timestamp,
    )
    return os.path.join(get_vote_data_dir(), filename)


def get_vote_data_files(name=None, datatype=None):

    vote_data_dir = get_vote_data_dir()

    files = []
    for filename in sorted(os.listdir(vote_data_dir)):

        if not os.path.isfile(os.path.join(vote_data_dir, filename)):
            continue

        parsed = parse_filename(filename)

        if name is not None and parsed['name'] != name:
            continue
        if datatype is not None and parsed['datatype'] != datatype:
            continue

        data = dict(parsed)
        data['path'] = os.path.join(vote_data_dir, filename)

        files.append(data)

    if datatype is not None:
        files = [file for file in files if file['datatype'] == datatype]

    return files


def parse_filename(filename):
    name, datatype, timestamp = os.path.splitext(filename)[0].split('__')
    return {
        'name': name,
        'datatype': datatype,
        'timestamp': timestamp,
    }


#
# # io
#


def load_raw_vote_data(
    name, datatype, request=None, timestamp=None, verbose=True
):
    """load vote data from disk, requesting it online if necessary"""

    path = None

    if request is not None and timestamp is not None:
        raise Exception('cannot specify both request and timestamp')

    elif timestamp is not None:
        # load specified file
        path = get_vote_data_path(
            name=name, datatype=datatype, timestamp=timestamp,
        )

    elif request is None and timestamp is None:
        # load most recent matching file
        files = get_vote_data_files(name=name, datatype=datatype)
        if len(files) > 0:
            path = files[-1]['path']

    # load data
    if path is not None:
        # load file
        if verbose:
            print('loading data path', path)
        with open(path, 'r') as f:
            return json.load(f)
    else:
        # request data
        return request_raw_vote_data(name=name, verbose=verbose)


#
# # crud
#


def get_vote_data(name, load_kwargs=None):

    # get raw data
    if load_kwargs is None:
        load_kwargs = {}
    load_kwargs = dict(load_kwargs)
    load_kwargs['name'] = name
    choices = load_raw_vote_data(datatype='choices', **load_kwargs)
    scores = load_raw_vote_data(datatype='scores', **load_kwargs)
    strategies = scores.keys()

    # process raw data
    votes = {}
    for address, raw in choices.items():
        address_scores = {
            strategy: float(scores[strategy][address])
            for strategy in strategies
        }

        # eliminate corrupted votes
        choice = raw['msg']['payload']['choice']
        if isinstance(choice, dict):
            continue

        address = address.lower()
        votes[address] = {
            'address': address,
            'timestamp': raw['msg']['timestamp'],
            'choice': choice,
            'scores': address_scores,
            'scores_total': sum(address_scores.values()),
        }

    return votes


def create_vote_dataframe(
    votes, tags_of_addresses=None, addresses_of_tags=None
):
    """

    ## Inputs
    - votes: vote data returned by get_vote_data()
    - addresses_of_tags: {tag_name: {address: tag_value}}
    - tags_of_addresses: {address: {tag_name: tag_value}}
    """

    votes = copy.deepcopy(votes)

    # create separate column for each voting strategy
    strategy_scores = {}
    for address, vote in votes.items():
        for strategy, strategy_score in vote['scores'].items():
            strategy_scores.setdefault(strategy, {})
            strategy_scores[strategy][address] = strategy_score
    if addresses_of_tags is None:
        addresses_of_tags = {}
    for strategy in strategy_scores:
        addresses_of_tags.setdefault(
            'score_' + strategy, strategy_scores[strategy]
        )

    # set tag values
    if tags_of_addresses is not None:
        for address in tags_of_addresses.keys():
            if address in votes:
                for tag_name, tag_value in tags_of_addresses.items():
                    votes[address][tag_name] = tag_value
    if addresses_of_tags is not None:
        for tag_name in addresses_of_tags.keys():
            for address, tag_value in addresses_of_tags[tag_name].items():
                if address in votes:
                    votes[address][tag_name] = tag_value

    votes_df = pandas.DataFrame(votes).transpose()

    # convert dtypes
    votes_df = votes_df.convert_dtypes()
    votes_df = votes_df.replace(np.nan, 0)

    return votes_df




from ctc import spec


def digest_shh_version(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_shh_post(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_shh_new_identity(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_shh_has_identity(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_shh_new_group(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_shh_add_to_group(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_shh_new_filter(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_shh_uninstall_filter(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_shh_get_filter_changes(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_shh_get_messages(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response




import pytest

from ctc import evm


@pytest.mark.asyncio
async def test_get_block():
    latest_block = await evm.async_get_block('latest')
    assert latest_block is not None


@pytest.mark.asyncio
async def test_fetch_latest_block_number():
    latest_block_number = await evm.async_get_latest_block_number()
    assert isinstance(latest_block_number, int)


@pytest.mark.asyncio
async def test_get_contract_creation_block():
    contract = '0x6B175474E89094C44Da98b954EedeAC495271d0F'
    result = await evm.async_get_contract_creation_block(contract)
    target = 8928158
    assert result == target




"""
# TODO
- amount burned/issued
    - need to sum up
        - minter rewards (both pre and post 1559)
        - uncle inclusion rewards
        - uncle mining rewards
    - see https://github.com/ethereum/go-ethereum/issues/1644
- amount of time in each block aggregation range
- historical blocks and block ranges
- color
"""

import time

import numpy as np
import pandas as pd
import toolstr
import tooltime

from ctc import evm
from ctc import rpc
from ctc.cli import cli_utils


def get_command_spec():
    return {
        'f': async_gas_command,
        'args': [
            {'name': '--last', 'kwargs': {'nargs': '+'}},
            {'name': '--output', 'kwargs': {'default': 'stdout'}},
            {'name': '--overwrite', 'kwargs': {'action': 'store_true'}},
        ],
    }


async def async_gas_command(last, output, overwrite):
    if last is None:
        last = [1, 10, 100]
    else:
        last = [
            int(subtoken.strip(' '))
            for token in last
            for subtoken in token.strip(',').split(',')
        ]

    n_blocks = max(last)
    latest = await rpc.async_eth_block_number()
    block_numbers = list(range(latest - n_blocks + 1, latest + 1))

    # get block transaction data
    blocks = await rpc.async_batch_eth_get_block_by_number(
        block_numbers=block_numbers,
        provider={'chunk_size': 1},
    )

    # compute block gas stats
    blocks_gas_stats = [evm.get_block_gas_stats(block) for block in blocks]
    gas_stats_df = pd.DataFrame(blocks_gas_stats, index=block_numbers)
    gas_stats_df.index.name = 'block'

    now = time.time()
    last_times = ['5 minutes', '10 minutes']
    last_times_blocks = []
    for last_time in last_times:
        cutoff_timestamp = now - tooltime.timelength_to_seconds(last_time)
        for i in range(n_blocks):
            if blocks[-i - 1]['timestamp'] < cutoff_timestamp:
                break
        print(i)
        last_times_blocks.append(i)

    toolstr.print_text_box('Gas Price Summary')
    print()
    print('all prices reported in gwei')
    print()
    print()
    toolstr.print_header('Latest block = ' + str(latest))
    for key, value in blocks_gas_stats[-1].items():
        if key in ['gas_limit', 'gas_used']:
            print(
                '-', key + ':', toolstr.format(value, order_of_magnitude=True)
            )
        else:
            print('-', key + ':', toolstr.format(value))
    print()
    print()
    toolstr.print_header('Previous Blocks')
    print('aggregated using median price of transactions in each block')
    print()
    headers = ['blocks', 'time', 'min', 'median', 'mean', 'max']
    rows = []
    for l, last_n in enumerate(last + last_times_blocks):

        row = []

        if l < len(last):
            if last_n == 1:
                row.append('latest block')
            else:
                row.append('last ' + str(last_n) + ' blocks')
            timelength_seconds = round(now - blocks[-last_n]['timestamp'])
            timelength_clock = tooltime.timelength_to_clock(timelength_seconds)
            row.append(timelength_clock)
        else:
            row.append('last ' + last_times[l - len(last)])
            row.append(tooltime.timelength_to_clock(last_times[l - len(last)]))

        sub_gas_stats_df = gas_stats_df.iloc[-last_n:]
        median_prices = sub_gas_stats_df['median_gas_price'].values
        median_prices = median_prices[~np.isnan(median_prices)]

        if len(median_prices) > 0:
            row.append(median_prices.min())
            row.append(np.median(median_prices))
            row.append(median_prices.mean())
            row.append(median_prices.max())
        else:
            row.extend([None] * 4)

        rows.append(row)

    final_df = pd.DataFrame(rows)
    final_df.columns = headers
    final_df = final_df.sort_values(by='time')
    final_df = final_df.set_index('blocks')

    cli_utils.output_data(final_df, output, overwrite)

    await rpc.async_close_http_session()




import typing

from ctc import binary
from ctc import spec
from .. import rpc_request
from .. import rpc_format


def construct_eth_call(
    to_address: spec.BinaryData,
    from_address: spec.BinaryData = None,
    gas: spec.BinaryData = None,
    gas_price: spec.BinaryData = None,
    value_sent: spec.BinaryData = None,
    block_number: spec.BlockSpec = None,
    call_data: spec.BinaryData = None,
    function_parameters: typing.Optional[typing.Union[list, dict]] = None,
    function_abi: typing.Optional[spec.FunctionABI] = None,
) -> spec.RpcResponse:

    if block_number is None:
        block_number = 'latest'
    block_number = rpc_format.encode_block_number(block_number)

    # encode call data
    if call_data is None:
        call_data = binary.encode_call_data(
            parameters=function_parameters,
            function_abi=function_abi,
        )

    # assemble request data
    call_object = {
        'to': to_address,
        'data': call_data,
        'from': from_address,
        'gas': gas,
        'gasPrice': gas_price,
        'value': value_sent,
    }
    call_object = {k: v for k, v in call_object.items() if v is not None}

    return rpc_request.create('eth_call', [call_object, block_number])


def construct_eth_estimate_gas(
    to_address: spec.BinaryData,
    from_address: spec.BinaryData = None,
    gas: spec.BinaryData = None,
    gas_price: spec.BinaryData = None,
    value_sent: spec.BinaryData = None,
    call_data: spec.BinaryData = None,
    function_parameters: typing.Optional[typing.Union[list, dict]] = None,
    function_abi: typing.Optional[spec.FunctionABI] = None,
) -> spec.RpcResponse:

    # encode call data
    if call_data is None:
        call_data = binary.encode_call_data(
            parameters=function_parameters,
            function_abi=function_abi,
        )

    # assemble call data
    call_object = {
        'to': to_address,
        'data': call_data,
        'from': from_address,
        'gas': gas,
        'gasPrice': gas_price,
        'value': value_sent,
    }
    call_object = {k: v for k, v in call_object.items() if v is not None}

    return rpc_request.create('eth_estimateGas', [call_object])


def construct_eth_get_balance(
    address,
    block_number=None,
) -> spec.RpcResponse:

    if block_number is None:
        block_number = 'latest'

    block_number = rpc_format.encode_block_number(block_number)
    return rpc_request.create('eth_getBalance', [address, block_number])


def construct_eth_get_storage_at(
    address: spec.BinaryData,
    position: spec.BinaryData,
    block_number='latest',
) -> spec.RpcResponse:

    position = binary.convert(position, 'prefix_hex')
    block_number = rpc_format.encode_block_number(block_number)
    return rpc_request.create(
        'eth_getStorageAt', [address, position, block_number]
    )


def construct_eth_get_code(
    address: spec.BinaryData,
    block_number: spec.BlockSpec = 'latest',
) -> spec.RpcResponse:

    block_number = rpc_format.encode_block_number(block_number)
    return rpc_request.create('eth_getCode', [address, block_number])




from ctc import evm
from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


async def async_eth_call(
    to_address,
    from_address=None,
    gas=None,
    gas_price=None,
    value_sent=None,
    block_number=None,
    call_data=None,
    function_parameters=None,
    provider=None,
    decode_response=True,
    delist_single_outputs=True,
    package_named_outputs=False,
    fill_empty=False,
    empty_token=None,
    function_abi=None,
    **function_abi_query
):

    if function_abi is None:
        function_abi = await evm.async_get_function_abi(
            contract_address=to_address, **function_abi_query
        )

    # construct request
    request = rpc_constructors.construct_eth_call(
        to_address=to_address,
        from_address=from_address,
        gas=gas,
        gas_price=gas_price,
        value_sent=value_sent,
        block_number=block_number,
        call_data=call_data,
        function_parameters=function_parameters,
        function_abi=function_abi,
    )

    # make request
    response = await rpc_request.async_send(request, provider=provider)

    # digest response
    return rpc_digestors.digest_eth_call(
        response,
        function_abi=function_abi,
        decode_response=decode_response,
        delist_single_outputs=delist_single_outputs,
        package_named_outputs=package_named_outputs,
        fill_empty=fill_empty,
        empty_token=empty_token,
    )


async def async_eth_estimate_gas(
    to_address,
    from_address=None,
    gas=None,
    gas_price=None,
    value_sent=None,
    call_data=None,
    function_parameters=None,
    provider=None,
    decode_response=True,
    function_abi=None,
    **function_abi_query
):

    if function_abi is None:
        function_abi = await evm.async_get_function_abi(
            contract_address=to_address, **function_abi_query
        )

    request = rpc_constructors.construct_eth_estimate_gas(
        to_address=to_address,
        from_address=from_address,
        gas=gas,
        gas_price=gas_price,
        value_sent=value_sent,
        call_data=call_data,
        function_parameters=function_parameters,
        function_abi=function_abi,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_estimate_gas(
        response,
        decode_response=decode_response,
    )


async def async_eth_get_balance(
    address,
    block_number='latest',
    provider=None,
    decode_response=True,
):
    request = rpc_constructors.construct_eth_get_balance(
        address=address,
        block_number=block_number,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_balance(
        response,
        decode_response=decode_response,
    )


async def async_eth_get_storage_at(
    address,
    position,
    block_number='latest',
    provider=None,
):
    request = rpc_constructors.construct_eth_get_storage_at(
        address=address,
        position=position,
        block_number=block_number,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_storage_at(response)


async def async_eth_get_code(address, block_number='latest', provider=None):
    request = rpc_constructors.construct_eth_get_code(
        address=address,
        block_number=block_number,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_code(response)




from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from typing_extensions import TypeGuard

from ctc import binary
from ..typedefs import block_types
from . import binary_typeguards


def is_block_number_name(
    block: typing.Any,
) -> TypeGuard[block_types.BlockNumberName]:
    return block in block_types.block_number_names


def is_raw_block_number(
    block: typing.Any,
) -> TypeGuard[block_types.RawBlockNumber]:

    # python3.7 compatibility
    # supports_int = isinstance(block, typing.SupportsInt)
    supports_int = hasattr(block, '__int__')

    return supports_int or binary_typeguards.is_hex_data(block)


def is_standard_block_number(
    block: typing.Any,
) -> TypeGuard[block_types.StandardBlockNumber]:
    return isinstance(block, int) or is_block_number_name(block)


def is_block_number_reference(
    block: typing.Any,
) -> TypeGuard[block_types.BlockNumberReference]:
    return is_raw_block_number(block) or is_standard_block_number(block)


def is_block_hash(block: typing.Any) -> TypeGuard[block_types.BlockHash]:
    return (
        binary_typeguards.is_hex_data(block)
        and binary.get_binary_n_bytes(block) == 32
    )


def is_block_reference(
    block: typing.Any,
) -> TypeGuard[block_types.BlockReference]:
    return is_block_number_reference(block) or is_block_hash(block)




from __future__ import annotations

import asyncio
import typing

from ctc import evm
from ctc import spec
from ctc import rpc
from . import coracle_spec


#
# # oracle getters
#


async def async_get_token_oracle(
    token: spec.Address,
    block: spec.BlockNumberReference = 'latest',
    provider: spec.ProviderSpec = None,
    replace_missing: bool = True,
    raise_if_missing: bool = True,
) -> spec.ContractAddress:

    coracle = coracle_spec.get_coracle_address(block=block)
    oracle = await rpc.async_eth_call(
        to_address=coracle,
        function_name='tokenToOracle',
        function_parameters=[token],
        block_number=block,
        provider=provider,
    )

    if replace_missing:
        oracle = await _async_replace_missing_oracle(
            token=token,
            oracle=oracle,
            provider=provider,
            replacement_block='latest',
        )

    if raise_if_missing:
        _ensure_oracle_valid(oracle, token, block)

    return oracle


async def async_get_tokens_oracles(
    tokens: typing.Sequence[spec.Address],
    block: spec.BlockNumberReference = 'latest',
    provider: spec.ProviderSpec = None,
    replace_missing: bool = True,
    raise_if_missing: bool = True,
) -> list[spec.ContractAddress]:

    block = await evm.async_block_number_to_int(block, provider=provider)
    oracles: typing.Sequence[
        spec.ContractAddress
    ] = await rpc.async_batch_eth_call(
        to_address=coracle_spec.get_coracle_address(block=block),
        function_name='tokenToOracle',
        function_parameter_list=[[token] for token in tokens],
        block_number=block,
        provider=provider,
    )

    if replace_missing:
        oracles = await _async_replace_missing_oracles(
            oracles=oracles,
            tokens=tokens,
            provider=provider,
            replacement_block=block,
        )

    if raise_if_missing:
        for oracle, token in zip(oracles, tokens):
            _ensure_oracle_valid(oracle, token, block)

    return list(oracles)


async def async_get_token_oracle_by_block(
    token: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    provider: spec.ProviderSpec = None,
) -> list[spec.ContractAddress]:

    coroutines = []
    for block in blocks:
        coroutine = async_get_token_oracle(
            token=token,
            block=block,
            provider=provider,
        )
        coroutines.append(coroutine)
    return await asyncio.gather(*coroutines)


#
# # price getters
#


async def async_get_token_price(
    token: spec.Address,
    block: spec.BlockReference = 'latest',
    provider: spec.ProviderSpec = None,
    normalize: bool = True,
) -> typing.Union[int, float]:

    # get oracle
    block = await evm.async_block_number_to_int(block, provider=provider)
    oracle = await async_get_token_oracle(
        token=token, block=block, provider=provider
    )

    # get price
    result = await rpc.async_eth_call(
        to_address=oracle,
        function_name='read',
        block_number=block,
        provider=provider,
    )
    price = result[0][0]

    if normalize:
        price /= 1e18

    return price


async def async_get_token_price_by_block(
    token: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    provider: spec.ProviderSpec = None,
    normalize: bool = True,
) -> typing.Union[list[int], list[float]]:

    # need to use c
    coroutines = []
    for block in blocks:
        coroutine = async_get_token_price(
            token=token,
            block=block,
            provider=provider,
            normalize=normalize,
        )
        coroutines.append(coroutine)
    return await asyncio.gather(*coroutines)


async def async_get_tokens_prices(
    tokens: typing.Sequence[spec.Address],
    block: spec.BlockReference = 'latest',
    provider: spec.ProviderSpec = None,
    normalize: bool = True,
) -> typing.Union[list[int], list[float]]:

    block = await evm.async_block_number_to_int(block, provider=provider)
    oracles = await async_get_tokens_oracles(
        tokens=tokens,
        block=block,
        provider=provider,
    )
    results = await rpc.async_batch_eth_call(
        to_addresses=oracles,
        function_name='read',
        block_number=block,
        provider=provider,
    )
    prices = [result[0][0] for result in results]

    if normalize:
        prices = [price / 1e18 for price in prices]

    return prices


#
# # fix broken onchain data
#


def _ensure_oracle_valid(
    oracle: spec.ContractAddress,
    token: spec.Address,
    block: spec.BlockNumberReference,
) -> None:
    if oracle == '0x0000000000000000000000000000000000000000':
        raise spec.MissingOracleException(
            'invalid oracle '
            + str(oracle)
            + ' for token '
            + str(token)
            + ' at block '
            + str(block)
        )


async def _async_replace_missing_oracle(
    oracle: spec.ContractAddress,
    token: spec.Address,
    provider: spec.ProviderSpec,
    replacement_block: spec.BlockNumberReference,
) -> spec.Address:

    if oracle == '0x0000000000000000000000000000000000000000':
        return await async_get_token_oracle(
            token=token,
            block=replacement_block,
            provider=provider,
            replace_missing=False,
            raise_if_missing=False,
        )
    else:
        return oracle


async def _async_replace_missing_oracles(
    oracles: typing.Sequence[spec.ContractAddress],
    tokens: typing.Sequence[spec.Address],
    provider: spec.ProviderSpec,
    replacement_block: spec.BlockNumberReference,
) -> typing.Sequence[spec.Address]:
    missing = [
        [oracle, token]
        for oracle, token in zip(oracles, tokens)
        if oracle == '0x0000000000000000000000000000000000000000'
    ]
    if len(missing) > 0:
        missing_oracles, missing_tokens = zip(*missing)
        results = await async_get_tokens_oracles(
            tokens=tokens,
            block=replacement_block,
            provider=provider,
            replace_missing=False,
            raise_if_missing=False,
        )
        result_iter = iter(results)
        replaced = []
        for oracle in oracles:
            if oracle == '0x0000000000000000000000000000000000000000':
                replaced.append(next(result_iter))
            else:
                replaced.append(oracle)
        return replaced

    else:
        return oracles

